#once

#include "library.s"

; PCD8544 Nokia 5110 LCD library
; LCD_Init              | Initialize
; LCD_Clear             | Clear LCD
; LCD_SetX              | Set X = FA16 0~83
; LCD_SetY              | Set Y = FA16 0~5
; LCD_SetPos            | Set character position FA16 = 0~13
; LCD_WriteChar         | Write Ascii Character FA16
; LCD_WriteStrF         | Write String in Flash at address FA32
; LCD_WriteDelayedStrF  | Write String in Flash at FA32 with cycles delay FB16

; GFX_Clear             | Clears internar video buffer
; GFX_Update            | Write the content of the video buffer to LCD
; GFX_PixelSet          | Set pixel at (FA,FB) (0~83, 0~47)
; GFX_Line              | Draw line (FA,FB) to (FC,FD)

;;;;;;;;;;;;;;;;;;;;;;;;
; Hardware connection:
; DATA      -> LCD DATA
; RAM2 CLK  -> LCD CLK
; OP bit 6  -> LCD D/C
; OP bit 7  -> LCD CS
;;;;;;;;;;;;;;;;;;;;;;;;

LCD_WIDTH   = 84
LCD_HEIGHT  = 48
LCD_SIZE    = LCD_WIDTH*LCD_HEIGHT/8

; Place GFX buffer at the end of SRAM, we do need about 504 bytes
GFX_BUFFER  = 0xFC00 ; The last 1kB

#ruledef
{
    ; Set X = 0~83, Y = 0~5
    LCD_SetXY {posx: u8}, {posy: u8} => asm {
        OUTI rvs8(0b10000000 | (posx & 0x7F))
        OUTI rvs8(0b01000000 | (posy & 0x07))
    }
    ; Set character position X = 0~13, Y = 0~5
    LCD_SetPos {posx: u8}, {posy: u8} => asm {
        OUTI rvs8(0b10000000 | ((posx*6 + 1) & 0x7F))
        OUTI rvs8(0b01000000 | (posy & 0x07))
    }
    LCD_WriteChar {cr: u8} => asm {
        LDI FA, cr
        CALL LCD_WriteChar
    }
    LCD_WriteStrF {str: u24} => asm {
        LDI FB, 0
        LDIW FA, str
        CALL LCD_WriteDelayedStrF
    }
    LCD_WriteDelayedStrF {cycles: u16}, {str: u24} => asm {
        LDIW FA, str
        LDI FB, cycles
        CALL LCD_WriteDelayedStrF
    }
}

LCD_Init:
    OUTI rvs8(0b00100001)       ; Set Extended function set
    OUTI rvs8(0b00000100 | 1)   ; Set T coefficient
    OUTI rvs8(0b00010000 | 4)   ; Set bias 0~4
    OUTI rvs8(0b10000000 | 60)  ; Set contrast 0~127
    OUTI rvs8(0b00100000)       ; Set Basic function set
    OUTI rvs8(0b00001100)       ; Set Normal mode
    OUTI rvs8(0b01000000)       ; Set Y = 0
    OUTI rvs8(0b10000000)       ; Set X = 0
    RET

LCD_Clear:
    LCD_SetXY 0, 0
    LDI FA, LCD_SIZE/4-1
.loop:
    OUT_LCD FA+1
    OUT_LCD FA+1
    OUT_LCD FA+1
    OUT_LCD FA+1
    DEC FA
    BRFC .loop
    RET

; X = FA16 0~83
LCD_SetX:
    ANDI FA, 0x7F
    ORI  FA, 0b10000000
    REV FA
    OUT FA + 1
    RET

; Y = FA16 0~5
LCD_SetY:
    ANDI FA, 0x07
    ORI  FA, 0b01000000
    REV FA
    OUT FA + 1
    RET

; Set Position X = FA16 0~13
LCD_SetPos:
    PUSH_PC
    LDI FB, 6
    CALL Multiply16
    ADDI FA, 1
    CALL LCD_SetX
    POP_PC
    RET

; Write Ascii Character FA16
LCD_WriteChar:
    PUSH_PC
    SUBI FA, 32
    BRFS .unknown
    BRSHI FA, 96, .unknown
    LDI FB, 5
    CALL Multiply16
    LDIW FB, LCD_Font
    ADD32_16 FB, FA
    LPM FA, FB
    OUT_LCD FA
    OUT_LCD FA+1
    ADDI32 FB, 2
    LPM FA, FB
    OUT_LCD FA
    OUT_LCD FA+1
    ADDI32 FB, 2
    LPB FA, FB
    OUT_LCD FA
    OUT_LCD FA+1
    POP_PC
    RET
.unknown:
    LDI FA, 0x55AA
    OUT_LCD FA
    OUT_LCD FA+1
    OUT_LCD FA
    OUT_LCD FA+1
    OUT_LCD FA
    LDI FA, 0
    OUT_LCD FA
    POP_PC
    RET

; Write String in Flash
; input: FA32 <- Address of the string in Flash
; input: FB16 <- Delay cycles
LCD_WriteDelayedStrF:
    PUSH_PC             ; Save return address
    PUSHW RA            ; Save RA
    PUSH RB
    MOVW RA, FA
    MOV RB, FB
.loop:
    LPB FA, RA          ; Load character from Flash
    BREQ FA, .stop      ; Test "\0" character
    REV FA              ; From MSB-first to LSB-first
    ANDI FA, FA+1, 0xFF ; Cast to 8bits
    CALL LCD_WriteChar  ; Write character
    ADDI32 RA, 1        ; Increase 32-bit pointer
    MOV TMP, RB         ; Load delay
    DelayLoop TMP       ; Insert delay loop
    JMP .loop
.stop:
    POP RB              ; Restore RB
    POPW RA             ; Restore RA
    POP_PC              ; Restore return address
    RET

LCD_WriteStrF:
    LDI FB, 0
    JMP LCD_WriteDelayedStrF

LCD_Font:
    #d 0x00, 0x00, 0x00, 0x00, 0x00 ; (space)
    #d 0x00, 0x00, 0x5F, 0x00, 0x00 ; !
    #d 0x00, 0x07, 0x00, 0x07, 0x00 ; "
    #d 0x14, 0x7F, 0x14, 0x7F, 0x14 ; #
    #d 0x24, 0x2A, 0x7F, 0x2A, 0x12 ; $
    #d 0x23, 0x13, 0x08, 0x64, 0x62 ; %
    #d 0x36, 0x49, 0x55, 0x22, 0x50 ; &
    #d 0x00, 0x05, 0x03, 0x00, 0x00 ; '
    #d 0x00, 0x1C, 0x22, 0x41, 0x00 ; (
    #d 0x00, 0x41, 0x22, 0x1C, 0x00 ; )
    #d 0x08, 0x2A, 0x1C, 0x2A, 0x08 ; *
    #d 0x08, 0x08, 0x3E, 0x08, 0x08 ; +
    #d 0x00, 0x50, 0x30, 0x00, 0x00 ; ,
    #d 0x08, 0x08, 0x08, 0x08, 0x08 ; -
    #d 0x00, 0x30, 0x30, 0x00, 0x00 ; .
    #d 0x20, 0x10, 0x08, 0x04, 0x02 ; /
    #d 0x3E, 0x51, 0x49, 0x45, 0x3E ; 0
    #d 0x00, 0x42, 0x7F, 0x40, 0x00 ; 1
    #d 0x42, 0x61, 0x51, 0x49, 0x46 ; 2
    #d 0x21, 0x41, 0x45, 0x4B, 0x31 ; 3
    #d 0x18, 0x14, 0x12, 0x7F, 0x10 ; 4
    #d 0x27, 0x45, 0x45, 0x45, 0x39 ; 5
    #d 0x3C, 0x4A, 0x49, 0x49, 0x30 ; 6
    #d 0x01, 0x71, 0x09, 0x05, 0x03 ; 7
    #d 0x36, 0x49, 0x49, 0x49, 0x36 ; 8
    #d 0x06, 0x49, 0x49, 0x29, 0x1E ; 9
    #d 0x00, 0x36, 0x36, 0x00, 0x00 ; :
    #d 0x00, 0x56, 0x36, 0x00, 0x00 ; ;
    #d 0x00, 0x08, 0x14, 0x22, 0x41 ; <
    #d 0x14, 0x14, 0x14, 0x14, 0x14 ; =
    #d 0x41, 0x22, 0x14, 0x08, 0x00 ; >
    #d 0x02, 0x01, 0x51, 0x09, 0x06 ; ?
    #d 0x32, 0x49, 0x79, 0x41, 0x3E ; @
    #d 0x7E, 0x11, 0x11, 0x11, 0x7E ; A
    #d 0x7F, 0x49, 0x49, 0x49, 0x36 ; B
    #d 0x3E, 0x41, 0x41, 0x41, 0x22 ; C
    #d 0x7F, 0x41, 0x41, 0x22, 0x1C ; D
    #d 0x7F, 0x49, 0x49, 0x49, 0x41 ; E
    #d 0x7F, 0x09, 0x09, 0x01, 0x01 ; F
    #d 0x3E, 0x41, 0x41, 0x51, 0x32 ; G
    #d 0x7F, 0x08, 0x08, 0x08, 0x7F ; H
    #d 0x00, 0x41, 0x7F, 0x41, 0x00 ; I
    #d 0x20, 0x40, 0x41, 0x3F, 0x01 ; J
    #d 0x7F, 0x08, 0x14, 0x22, 0x41 ; K
    #d 0x7F, 0x40, 0x40, 0x40, 0x40 ; L
    #d 0x7F, 0x02, 0x04, 0x02, 0x7F ; M
    #d 0x7F, 0x04, 0x08, 0x10, 0x7F ; N
    #d 0x3E, 0x41, 0x41, 0x41, 0x3E ; O
    #d 0x7F, 0x09, 0x09, 0x09, 0x06 ; P
    #d 0x3E, 0x41, 0x51, 0x21, 0x5E ; Q
    #d 0x7F, 0x09, 0x19, 0x29, 0x46 ; R
    #d 0x46, 0x49, 0x49, 0x49, 0x31 ; S
    #d 0x01, 0x01, 0x7F, 0x01, 0x01 ; T
    #d 0x3F, 0x40, 0x40, 0x40, 0x3F ; U
    #d 0x1F, 0x20, 0x40, 0x20, 0x1F ; V
    #d 0x7F, 0x20, 0x18, 0x20, 0x7F ; W
    #d 0x63, 0x14, 0x08, 0x14, 0x63 ; X
    #d 0x03, 0x04, 0x78, 0x04, 0x03 ; Y
    #d 0x61, 0x51, 0x49, 0x45, 0x43 ; Z
    #d 0x00, 0x00, 0x7F, 0x41, 0x41 ; [
    #d 0x02, 0x04, 0x08, 0x10, 0x20 ; "\"
    #d 0x41, 0x41, 0x7F, 0x00, 0x00 ; ]
    #d 0x04, 0x02, 0x01, 0x02, 0x04 ; ^
    #d 0x40, 0x40, 0x40, 0x40, 0x40 ; _
    #d 0x00, 0x01, 0x02, 0x04, 0x00 ; `
    #d 0x20, 0x54, 0x54, 0x54, 0x78 ; a
    #d 0x7F, 0x48, 0x44, 0x44, 0x38 ; b
    #d 0x38, 0x44, 0x44, 0x44, 0x20 ; c
    #d 0x38, 0x44, 0x44, 0x48, 0x7F ; d
    #d 0x38, 0x54, 0x54, 0x54, 0x18 ; e
    #d 0x08, 0x7E, 0x09, 0x01, 0x02 ; f
    #d 0x08, 0x14, 0x54, 0x54, 0x3C ; g
    #d 0x7F, 0x08, 0x04, 0x04, 0x78 ; h
    #d 0x00, 0x44, 0x7D, 0x40, 0x00 ; i
    #d 0x20, 0x40, 0x44, 0x3D, 0x00 ; j
    #d 0x00, 0x7F, 0x10, 0x28, 0x44 ; k
    #d 0x00, 0x41, 0x7F, 0x40, 0x00 ; l
    #d 0x7C, 0x04, 0x18, 0x04, 0x78 ; m
    #d 0x7C, 0x08, 0x04, 0x04, 0x78 ; n
    #d 0x38, 0x44, 0x44, 0x44, 0x38 ; o
    #d 0x7C, 0x14, 0x14, 0x14, 0x08 ; p
    #d 0x08, 0x14, 0x14, 0x18, 0x7C ; q
    #d 0x7C, 0x08, 0x04, 0x04, 0x08 ; r
    #d 0x48, 0x54, 0x54, 0x54, 0x20 ; s
    #d 0x04, 0x3F, 0x44, 0x40, 0x20 ; t
    #d 0x3C, 0x40, 0x40, 0x20, 0x7C ; u
    #d 0x1C, 0x20, 0x40, 0x20, 0x1C ; v
    #d 0x3C, 0x40, 0x30, 0x40, 0x3C ; w
    #d 0x44, 0x28, 0x10, 0x28, 0x44 ; x
    #d 0x0C, 0x50, 0x50, 0x50, 0x3C ; y
    #d 0x44, 0x64, 0x54, 0x4C, 0x44 ; z
    #d 0x00, 0x08, 0x36, 0x41, 0x00 ; {
    #d 0x00, 0x00, 0x7F, 0x00, 0x00 ; |
    #d 0x00, 0x41, 0x36, 0x08, 0x00 ; }
    #d 0x08, 0x08, 0x2A, 0x1C, 0x08 ; ->
    #d 0x08, 0x1C, 0x2A, 0x08, 0x08 ; <-

GFX_Clear:
    PUSH_PC
    LDI FA, GFX_BUFFER
    LDI FB, LCD_SIZE
    CALL MemClr
    POP_PC
    RET

GFX_Update:
    PUSH_PC
    LCD_SetXY 0, 0
    LDI FA, GFX_BUFFER
    LDI FB, LCD_SIZE-2
.loop:
    LD FC, FA
    OUT_LCD FC
    OUT_LCD FC+1
    ADDI FA, 2
    SUBI FB, 2
    BRFC .loop
    POP_PC
    RET

GFX_PixelSet:
    PUSH_PC
    ADDI FA, GFX_BUFFER
    ANDI TMP, FB, 0x08
    BRFS .skip1
    ADDI FA, LCD_WIDTH
.skip1:
    ANDI TMP, FB, 0x10
    BRFS .skip2
    ADDI FA, 2*LCD_WIDTH
.skip2:
    ANDI TMP, FB, 0x20
    BRFS .skip4
    ADDI FA, 4*LCD_WIDTH
.skip4:
    MOVW FC, FA
    ANDI FA, FB, 7
    CALL BinShiftRev8
    LDB FB, FC
    OR FB, FA
    STB FC, FB
    POP_PC
    RET

GFX_Line:
    PUSH_PC
    PUSH RA
    PUSH RB
    PUSH RC
    PUSH RD
    PUSH RE
    PUSH RF
    MOV RA, FA      ; start X
    MOV RB, FB      ; start Y
    LDI RC, 1       ; X positive
    LDI RD, 1       ; Y positive
    SUB FE, FC, FA
    BRFC .positive1
    LDI RC, -1      ; X negative
    NEG FE
.positive1:
    SUB FF, FD, FB
    BRFC .positive2
    LDI RD, -1      ; Y negative
    NEG FF
.positive2:
    BRLO FE, FF, .vertical
.horizontal:
    LSL FF          ; FE = x2, FF = C0
    LSL RE, FE
    SUB RE, FF, RE  ; C1
    SUB RF, FF, FE  ; p
.horizontal_loop:
    MOV FA, RA
    MOV FB, RB
    CALL GFX_PixelSet
    ADD RA, RC
    ANDI TMP, RF, 0x8000
    BRFS .horizontal_ppositive
    ADD RF, FF
    DEC FE
    BRFC .horizontal_loop
    JMP .stop
.horizontal_ppositive:
    ADD RF, RE
    ADD RB, RD
    DEC FE
    BRFC .horizontal_loop
    JMP .stop
.vertical:
    LSL FE          ; FE = C0, FF = y2
    LSL RE, FF
    SUB RE, FE, RE  ; C1
    SUB RF, FE, FF  ; p
.vertical_loop:
    MOV FA, RA
    MOV FB, RB
    CALL GFX_PixelSet
    ADD RB, RD
    ANDI TMP, RF, 0x8000
    BRFS .vertical_ppositive
    ADD RF, FE
    DEC FF
    BRFC .vertical_loop
    JMP .stop
.vertical_ppositive:
    ADD RF, RE
    ADD RA, RC
    DEC FF
    BRFC .vertical_loop
    JMP .stop
.stop:
    POP RF
    POP RE
    POP RD
    POP RC
    POP RB
    POP RA
    POP_PC
    RET
