#once

#include "library.s"

; Library for an Alphanumeric SPI LCD with unknown chipset
; LCD_Init              | Initialize
; LCD_Clear             | Clear LCD
; LCD_SetPosition       | Set position = FA16 0~32
; LCD_Write             | Write default charset FA16 0~512
; LCD_WriteAscii        | Write Ascii character FA16
; LCD_WriteStrF         | Write String in Flash at address FA32

; HP_LCD_Demo           | Run Demo

;;;;;;;;;;;;;;;;;;;;;;;;
; Hardware connection:
; DATA      -> LCD DATA
; RAM2 CLK  -> LCD CLK
; OP bit 6  -> LCD RESET
; OP bit 7  -> LCD CS
;;;;;;;;;;;;;;;;;;;;;;;;

; Place LCD cursor location at the end of RAM
LCD_Cursor = 0xFC00 ; The last 1kB

#ruledef
{
    ; Set pos = 0~32
    LCD_SetPosition {pos: u8} => asm {
        LDI LCD_Cursor, (pos&0x1F)
        OUTI rvs8(0x38)
        OUTI rvs8(0x80 | (pos&0xF) | ((pos&0x10)<<2))
        OUTI rvs8(0x3C)
    }
    LCD_Write {cr: u16} => asm {
        LDI FA, cr
        CALL LCD_Write
    }
    LCD_WriteAscii {cr: u16} => asm {
        LDI FA, cr
        CALL LCD_WriteAscii
    }
    LCD_WriteStrF {str: u24} => asm {
        LDIW FA, str
        CALL LCD_WriteStrF
    }
}

LCD_Init:
    LDI FA, 0
    OUT2 FA         ; Reset LCD
    DelayCycles delay8MHz_ms(2)
    OUTI rvs8(0x33) ; Set 8-bit interface
    OUTI rvs8(0x33) ; Set 8-bit interface
    OUTI rvs8(0x33) ; Set 8-bit interface
    OUTI rvs8(0x08) ; Display off
    OUTI rvs8(0x01) ; Clear
    DelayCycles delay8MHz_ms(2)
    OUTI rvs8(0x02) ; Home
    DelayCycles delay8MHz_ms(2)
    OUTI rvs8(0x3C) ; 2 Lines, Command entry
    OUTI rvs8(0x06) ; Power on
    OUTI rvs8(0x45) ; No Flip
    OUTI rvs8(0x0C) ; LCD on, no cursor, english charset
    DelayCycles delay8MHz_ms(2)
    LDI LCD_Cursor, 0
    RET

LCD_Clear:
    OUTI rvs8(0x01) ; Clear
    DelayCycles delay8MHz_ms(2)
    OUTI rvs8(0x02) ; Home
    DelayCycles delay8MHz_ms(2)
    RET

; Set position = FA16 0~32
LCD_SetPosition:
    ANDI LCD_Cursor, FA, 0x1F
    OUTI rvs8(0x38)
    BRLOI LCD_Cursor, 0x10, .skip
    ANDI FA, 0x0F
    ORI FA, 0xC0
    JMP .write
.skip:
    ORI FA, LCD_Cursor, 0x80
.write:
    REV FA
    OUT FA + 1
    OUTI rvs8(0x3C)
    RET

; Write default charset FA16 0~512
LCD_Write:
    ANDI TMP, FA, 0x100
    BRFS .default
    OUTI rvs8(0xD)  ; Special character set
    OUTI rvs8(0x80)
    REV FA
    OUT FA+1
    OUTI rvs8(0xC)  ; Default character set
    JMP .continue
.default:
    OUTI rvs8(0x80)
    REV FA
    OUT FA+1
.continue:
    INC LCD_Cursor
    ANDI TMP, LCD_Cursor, 0xF
    BRFC .stop
    PUSH_PC
    MOV FA, LCD_Cursor
    CALL LCD_SetPosition
    POP_PC
.stop:
    RET

; Write Ascii character FA16
LCD_WriteAscii:
    BRNEI FA, 32, .not_a_space
    JMP LCD_Write
.not_a_space:
    BRNEI FA, 42, .not_a_star
    LDI FA, 0x01FF
    JMP LCD_Write
.not_a_star:
    BRNEI FA, 115, .not_an_s
    LDI FA, 0x0100 + 143
    JMP LCD_Write
.not_an_s:
    SUBI FA, 10
    JMP LCD_Write

; Write String in Flash
; input: FA32 <- Address of the string in Flash
LCD_WriteStrF:
    PUSH_PC             ; Save return address
    PUSHW RA            ; Save RA
    MOVW RA, FA
.loop:
    LPB FA, RA          ; Load character from Flash
    BREQ FA, .stop      ; Test "\0" character
    REV FA              ; From MSB-first to LSB-first
    ANDI FA, FA+1, 0xFF ; Cast to 8bits
    CALL LCD_WriteAscii ; Write character
    ADDI32 RA, 1        ; Increase 32-bit pointer
    JMP .loop
.stop:
    POPW RA             ; Restore RA
    POP_PC              ; Restore return address
    RET

HP_LCD_Str1:
    #d "Hello World!\0"
HP_LCD_Str2:
    #d "from: 16-bit CPU\0"
HP_LCD_Demo:
    PUSH_PC
    PUSH RA
    PUSH RB
    CALL LCD_Init

    LCD_SetPosition 2
    LCD_WriteStrF HP_LCD_Str1
    LCD_SetPosition 16
    LCD_WriteStrF HP_LCD_Str2

    LDI RA, 0
    LCD_SetPosition 0
.loop:
    CALL WaitForButton
    LDI RB, 32-1
.inner_loop:
    MOV FA, RA
    CALL LCD_WriteAscii
    INC RA
    DEC RB
    BRFC .inner_loop
    DelayCycles delay8MHz_ms(250)
    ANDI RA, 0x1FF
    JMP .loop

    HALT
    POP RB
    POP RA
    POP_PC
    RET
