#once

; Library Functions
; Multiply16            | Returns FA16 = FA16 * FB16
; Multiply32_16x16      | Returns FA32 = FA16 * FB16
; Multiply32_16x16_signed | Returns signed FA32 = FA16 * FB16
; Multiply32_32x16      | Returns FA32 = FA32 * FB16
; Multiply64_64x16      | Returns WFA64 = WFA64 * FA16
; Divide16              | Returns FA16 = FA16 / FB16
; Divide32_32x16        | Returns FA32 = FA32 / FB16
; Divide32_32x16_signed | Returns FA32 = FA32 / FB16
; Divide32_32x32        | Returns FA32 = FA32 / FB32
; Divide64_64x64        | Returns WFA64 = WFA64 / WFB16

; BinShift8             | Returns FA16 = (1<<FA) 0~7
; BinShiftRev8          | Returns FA16 = (1<<(7-FA)) 0~7
; BinShift16            | Returns FA16 = (1<<FA) 0~15
; BinShiftRev16         | Returns FA16 = (1<<(15-FA)) 0~15

; MemSet                | Writes from addr in FA16 length FB16, contents of FC8
; MemClr                | Clears from addr in FA16 length FB16

; DelayLoop             | Delay cycles register
; DelayCycles u16       | Delay cycles immediate
; delay8MHz_ms(ms)      | Returns u16 with cycles at 8MHz
; delay10MHz_ms(ms)     | Returns u16 with cycles at 10MHz

; WaitForButton         | Loops until the input button is pressed

; Uses custom f16 fixed point format
; CosFP                 | Returns FAf16 = sin(FA16) degrees 0~360
; SinFP                 | Returns FAf16 = sin(FA16) degrees 0~360

;;;;;;;;;;;;;;;;;;;;;;;;


; Returns FA16 = FA16 * FB16
; FB should be smaller for faster run time
Multiply16:
    LDI FC, 0       ; Clear result
.loop:
    ANDI TMP, FB, 1
    BRFS .skip_add
    ADD FC, FA      ; Add FC += FA
.skip_add:
    LSL FA          ; Shift FA <<= 1
    LSR FB
    BRNE FB, .loop
    MOV FA, FC      ; Copy result
    RET

; Returns FA32 = FA32 * FB16
; FB should be smaller for faster run time
Multiply32_32x16:
    LDIW FC, 0      ; Clear result
.loop:
    ANDI TMP, FB, 1
    BRFS .skip_add
    ADD32 FC, FA    ; Add FC32 += FA32
.skip_add:
    LSL32 FA        ; Shift FA <<= 1
    LSR FB
    BRNE FB, .loop
    MOVW FA, FC     ; Copy result
    RET

; Returns FA32 = FA16 * FB16
Multiply32_16x16:
    LDI FA+2, 0     ; Cast FA16 to FA32
    JMP Multiply32_32x16

; Returns FA32 = FA16 * FB16
; FB should be smaller for faster run time
Multiply32_16x16_signed:
    PUSH_PC
    ANDI FE, FA, 0x8000
    BRFS .positive1
    NEG FA
.positive1:
    ANDI FF, FB, 0x8000
    BRFS .positive2
    NEG FB
.positive2:
    CALL Multiply32_16x16
    XOR TMP, FE, FF
    BRFS .stop
    NEG32 FA
.stop:
    POP_PC
    RET

; Returns WFA64 = WFA64 * FA16
Multiply64_64x16:
    LDIW WFB, 0     ; Clear result
    LDIW WFB+4, 0   ; Clear result
.loop:
    ANDI TMP, FA, 1
    BRFS .skip_add
    ADD32 WFB, WFA  ; Add
    ADC WFB+4, WFA+4
    ADC WFB+6, WFA+6
.skip_add:
    LSL32 WFA       ; Shift FA <<= 1
    ROL32 WFA+4
    LSR FA
    BRNE FA, .loop
    MOVW WFA, WFB   ; Copy result
    MOVW WFA+4, WFB+4
    RET

; Returns FA16 = FA16 / FB16
Divide16:
    LDI FC, 0
    LDI FD, 17-1    ; 17 cycles
    CLF
.loop:
    ROL FA
    ROL FC
    DEC FD
    BRFC .continue
    RET
.continue:
    SUB FC, FB      ; Test if we can subtract
    BRFC .subtract_ok
    ADD FC, FB
    CLF
    JMP .loop
.subtract_ok:
    STF
    JMP .loop

; Returns FA32 = FA32 / FB32
Divide32_32x32:
    LDIW FC, 0
    LDI FD, 33-1    ; 33 cycles
    CLF
.loop:
    ROL32 FA
    ROL32 FC
    DEC FD
    BRFC .continue
    RET
.continue:
    SUB32 FC, FB
    BRFC .subtract_ok
    ADD32 FC, FB
    CLF
    JMP .loop
.subtract_ok:
    STF
    JMP .loop

; Returns FA32 = FA32 / FB16
Divide32_32x16:
    LDI FB+2, 0     ; Cast FB16 to FB32
    JMP Divide32_32x32

; Returns FA32 = FA32 / FB16
Divide32_32x16_signed:
    PUSH_PC
    ANDI FE, FA+2, 0x8000
    BRFS .positive1
    NEG32 FA
.positive1:
    ANDI FF, FB, 0x8000
    BRFS .positive2
    NEG FB
.positive2:
    CALL Divide32_32x16
    XOR TMP, FE, FF
    BRFS .stop
    NEG32 FA
.stop:
    POP_PC
    RET

; Returns WFA64 = WFA64 / WFB16
Divide64_64x64:
    LDIW WFC, 0
    LDIW WFC+4, 0
    LDI FA, 65-1    ; 65 cycles
    CLF
.loop:
    ROL32 WFA
    ROL32 WFA+4
    ROL32 WFC
    ROL32 WFC+4
    DEC FA
    BRFC .continue
    RET
.continue:
    SUB32 WFC, WFB
    SBC WFC+4, WFB+4
    SBC WFC+6, WFB+6
    BRFC .subtract_ok
    ADD32 WFC, WFB
    ADC WFC+4, WFB+4
    ADC WFC+6, WFB+6
    CLF
    JMP .loop
.subtract_ok:
    STF
    JMP .loop

;Returns FA8 = (1<<FA) 0~7
BinShift8_lookup:
    #d 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
BinShiftRev8_lookup:
    #d 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80
BinShift8:
    LDI FA+2, 0
    ADDI32 FA, BinShift8_lookup
    LPB FA, FA
    RET
BinShiftRev8:
    LDI FA+2, 0
    ADDI32 FA, BinShiftRev8_lookup
    LPB FA, FA
    RET

;Returns FA16 = (1<<FA) 0~15
BinShift16_lookup:
    #d 0x8000, 0x4000, 0x2000, 0x1000, 0x0800, 0x0400, 0x0200, 0x0100
    #d 0x0080, 0x0040, 0x0020, 0x0010, 0x0008, 0x0004, 0x0002, 0x0001
BinShiftRev16_lookup:
    #d 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080
    #d 0x0100, 0x0200, 0x0400, 0x0800, 0x1000, 0x2000, 0x4000, 0x8000
BinShift16:
    LSL FA
    LDI FA+2, 0
    ADDI32 FA, BinShift16_lookup
    LPM FA, FA
    RET
BinShiftRev16:
    LSL FA
    LDI FA+2, 0
    ADDI32 FA, BinShiftRev16_lookup
    LPM FA, FA
    RET

; Writes from addr in FA16 length FB16, contents of FC8
MemSet:
    MOV FC+1, FC    ; Copy lower 8 bits
    SUBI FB, 2
    BRFS .odd
.loop:
    ST FA, FC       ; Write value
    ADDI FA, 2
    SUBI FB, 2
    BRFC .loop
.odd:
    ADDI FB, 1      ; Finished?
    BRFC .even
    STB FA, FC      ; Write last odd value
.even:
    RET

; Clears from addr in FA16 length FB16
MemClr:
    PUSH_PC
    LDI FC, 0       ; Write 0
    CALL MemSet
    POP_PC
    RET

#ruledef
{
    DelayLoop {cycles: reg} => asm {
        DEC cycles
        BRFC ($-5)
    }
    DelayCycles {cycles: u16} => asm {
        LDI TMP, cycles
        DelayLoop TMP
    }
}

#fn delay8MHz_ms(ms) =>
{
    ((ms*8000/(774+625))-1)&0xFFFF
}

#fn delay10MHz_ms(ms) =>
{
    ((ms*10000/(774+625))-1)&0xFFFF
}

WaitForButton:
    ANDI TMP, 0, 0x02
    BRFS WaitForButton
    RET

FP_DECIMAL = 14
FP_ONE     = 1<<FP_DECIMAL
FP_HALF    = 1<<(FP_DECIMAL-1)

COS_LOOKUP:
    #d16 rvs16(16384), rvs16(16381), rvs16(16374), rvs16(16361), rvs16(16344)
    #d16 rvs16(16321), rvs16(16294), rvs16(16261), rvs16(16224), rvs16(16182)
    #d16 rvs16(16135), rvs16(16082), rvs16(16025), rvs16(15964), rvs16(15897)
    #d16 rvs16(15825), rvs16(15749), rvs16(15668), rvs16(15582), rvs16(15491)
    #d16 rvs16(15395), rvs16(15295), rvs16(15190), rvs16(15081), rvs16(14967)
    #d16 rvs16(14848), rvs16(14725), rvs16(14598), rvs16(14466), rvs16(14329)
    #d16 rvs16(14188), rvs16(14043), rvs16(13894), rvs16(13740), rvs16(13582)
    #d16 rvs16(13420), rvs16(13254), rvs16(13084), rvs16(12910), rvs16(12732)
    #d16 rvs16(12550), rvs16(12365), rvs16(12175), rvs16(11982), rvs16(11785)
    #d16 rvs16(11585), rvs16(11381), rvs16(11173), rvs16(10963), rvs16(10748)
    #d16 rvs16(10531), rvs16(10310), rvs16(10086), rvs16(9860), rvs16(9630)
    #d16 rvs16(9397), rvs16(9161), rvs16(8923), rvs16(8682), rvs16(8438)
    #d16 rvs16(8191), rvs16(7943), rvs16(7691), rvs16(7438), rvs16(7182)
    #d16 rvs16(6924), rvs16(6663), rvs16(6401), rvs16(6137), rvs16(5871)
    #d16 rvs16(5603), rvs16(5334), rvs16(5062), rvs16(4790), rvs16(4516)
    #d16 rvs16(4240), rvs16(3963), rvs16(3685), rvs16(3406), rvs16(3126)
    #d16 rvs16(2845), rvs16(2563), rvs16(2280), rvs16(1996), rvs16(1712)
    #d16 rvs16(1427), rvs16(1142), rvs16(857), rvs16(571), rvs16(285)
    #d16 rvs16(0)

CosFP:
    LDIW FB, COS_LOOKUP
    BRSHI FA, 90, .larger_than_90
.load_positive:
    LSL FA
    ADD32_16 FB, FA
    LPM FA, FB
    RET
.larger_than_90:
    BRSHI FA, 180, .larger_than_180
    LDI FC, 180
    SUB FA, FC, FA
.load_negative:
    LSL FA
    ADD32_16 FB, FA
    LPM FA, FB
    NEG FA
    RET
.larger_than_180:
    BRSHI FA, 270, .larger_than_270
    SUBI FA, 180
    JMP .load_negative
.larger_than_270:
    LDI FC, 360
    SUB FA, FC, FA
    JMP .load_positive

SinFP:
    SUBI FA, 90
    BRFC CosFP
    ADDI FA, 360
    JMP CosFP
