#bits 8

FLASH_START     = 0x040000
FLASH_END       = 0x100000

#bankdef program
{
    #addr       FLASH_START
    #addr_end   FLASH_END
    #outp       0
}

; Named register addresses:
RAM_ADDR_FLAG   = 0x0000
RAM_ADDR_PC     = 0x0001
;RAM_ADDR_PTR    = 0x0004 ; Not accessible in RAM1, only in RAM2
;RAM_ADDR_VAL    = 0x0006 ; Not accessible in RAM1, only in RAM2
RAM_ADDR_SP     = 0x0008
RAM_ADDR_SPVAL  = 0x000A
RAM_ADDR_TMP    = 0x0010 ; TMP scratch register
RAM_ADDR_REG    = 0x0014 ; Start address for registers

; 32bit Temporary/scratch register
TMP = RAM_ADDR_TMP

; 32bit Registers, Function-saved
; Functions must save and restore these registers if used
RA  = RAM_ADDR_REG+0*4
RB  = RAM_ADDR_REG+1*4
RC  = RAM_ADDR_REG+2*4
RD  = RAM_ADDR_REG+3*4
RE  = RAM_ADDR_REG+4*4
RF  = RAM_ADDR_REG+5*4
RG  = RAM_ADDR_REG+6*4
RH  = RAM_ADDR_REG+7*4

; 32-bit Function arguments and return values, Function-clobbered
; Functions do not have to restore these registers and can freely use them
FA  = RAM_ADDR_REG+8*4
FB  = RAM_ADDR_REG+9*4
FC  = RAM_ADDR_REG+10*4
FD  = RAM_ADDR_REG+11*4
FE  = RAM_ADDR_REG+12*4
FF  = RAM_ADDR_REG+13*4
FG  = RAM_ADDR_REG+14*4
FH  = RAM_ADDR_REG+15*4

; 64-bit 'long long' registers, Function-saved
WA  = RAM_ADDR_REG+16*4
WB  = RAM_ADDR_REG+18*4
WC  = RAM_ADDR_REG+20*4
WD  = RAM_ADDR_REG+22*4

; 64-bit 'long long' registers, Function-clobbered
WFA = RAM_ADDR_REG+24*4
WFB = RAM_ADDR_REG+26*4
WFC = RAM_ADDR_REG+28*4
WFD = RAM_ADDR_REG+30*4

; User address space
RAM_ADDR_USER = 0x0100

#subruledef reg
{
    {r: u16} => r
    r{r: u16} => r
}

#fn rvs8(value) =>
{
    (value[0:0] @ value[1:1] @ value[2:2] @ value[3:3] @
     value[4:4] @ value[5:5] @ value[6:6] @ value[7:7])
}

#fn rvs16(value) =>
{
    (rvs8(value[7:0]) @ rvs8(value[15:8]))
}

#fn rvs24(value) =>
{
    (rvs16(value[15:0]) @ rvs8(value[23:16]))
}

#fn rvs32(value) =>
{
    (rvs16(value[15:0]) @ rvs16(value[31:16]))
}

#ruledef
{
    ; No operands
    INIT => 0x00
    RESET => 0x01
    STF  => 0x0E
    CLF  => 0x0F
    NOP  => 0x10
    HALT => 0x3F
    ; Aliases
    RST  => asm { RESET }
    HLT  => asm { HALT }
    
    ; Reg <- Reg
    MOV  {dst: reg}, {src: reg} => 0x11 @ dst @ src
    MOVW {dst: reg}, {src: reg} => 0x12 @ dst @ src
    INC  {dst: reg}, {src: reg} => 0x13 @ dst @ src
    DEC  {dst: reg}, {src: reg} => 0x14 @ dst @ src
    COM  {dst: reg}, {src: reg} => 0x15 @ dst @ src
    NEG  {dst: reg}, {src: reg} => 0x16 @ dst @ src
    LSL  {dst: reg}, {src: reg} => 0x17 @ dst @ src
    LSR  {dst: reg}, {src: reg} => 0x18 @ dst @ src
    ROL  {dst: reg}, {src: reg} => 0x19 @ dst @ src
    ROR  {dst: reg}, {src: reg} => 0x1A @ dst @ src
    ASR  {dst: reg}, {src: reg} => 0x1B @ dst @ src
    REV  {dst: reg}, {src: reg} => 0x1C @ dst @ src
    
    ; In place
    INC  {dst: reg} => asm { INC dst, dst }
    DEC  {dst: reg} => asm { DEC dst, dst }
    COM  {dst: reg} => asm { COM dst, dst }
    NEG  {dst: reg} => asm { NEG dst, dst }
    LSL  {dst: reg} => asm { LSL dst, dst }
    LSR  {dst: reg} => asm { LSR dst, dst }
    ROL  {dst: reg} => asm { ROL dst, dst }
    ROR  {dst: reg} => asm { ROR dst, dst }
    ASR  {dst: reg} => asm { ASR dst, dst }
    REV  {dst: reg} => asm { REV dst, dst }
    
    ; Wide types
    LSL32 {dst: reg}, {src: reg} => asm {
        LSL dst, src
        ROL dst+2, src+2
    }
    ROL32 {dst: reg}, {src: reg} => asm {
        ROL dst, src
        ROL dst+2, src+2
    }
    NEG32 {dst: reg}, {src: reg} => asm {
        COM dst+2, src+2
        NEG dst, src
        ADCI dst+2, 0
    }
    REV32 {dst: reg}, {src: reg} => asm {
        REV dst, src+2
        REV dst+2, src
    }
    LSL32 {dst: reg} => asm { LSL32 dst, dst }
    ROL32 {dst: reg} => asm { ROL32 dst, dst }
    NEG32 {dst: reg} => asm { NEG32 dst, dst }
    
    ; Reg <- Reg + Value
    ADDI {dst: reg}, {src: reg}, {val: i16} => 0x1D @ dst @ src @ rvs16(val)
    ADCI {dst: reg}, {src: reg}, {val: i16} => 0x1E @ dst @ src @ rvs16(val)
    SUBI {dst: reg}, {src: reg}, {val: i16} => 0x1F @ dst @ src @ rvs16(val)
    SBCI {dst: reg}, {src: reg}, {val: i16} => 0x20 @ dst @ src @ rvs16(val)
    ANDI {dst: reg}, {src: reg}, {val: i16} => 0x21 @ dst @ src @ rvs16(val)
    ORI  {dst: reg}, {src: reg}, {val: i16} => 0x22 @ dst @ src @ rvs16(val)
    XORI {dst: reg}, {src: reg}, {val: i16} => 0x23 @ dst @ src @ rvs16(val)
    
    ; In place
    ADDI {dst: reg}, {val: i16} => asm { ADDI dst, dst, val }
    ADCI {dst: reg}, {val: i16} => asm { ADCI dst, dst, val }
    SUBI {dst: reg}, {val: i16} => asm { SUBI dst, dst, val }
    SBCI {dst: reg}, {val: i16} => asm { SBCI dst, dst, val }
    ANDI {dst: reg}, {val: i16} => asm { ANDI dst, dst, val }
    ORI  {dst: reg}, {val: i16} => asm {  ORI dst, dst, val }
    XORI {dst: reg}, {val: i16} => asm { XORI dst, dst, val }
    
    ; Wide types
    ADDI32 {dst: reg}, {src: reg}, {val: i32} => asm {
        ADDI dst, src, (val&0xFFFF)
        ADCI dst+2, src+2, (val>>16)
    }
    ADDI32 {dst: reg}, {val: i32} => asm { ADDI32 dst, dst, val }
    
    
    ; Reg <- Reg + Reg
    ADD  {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x24
    ADC  {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x25
    SUB  {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x26
    SBC  {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x27
    AND  {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x28
    OR   {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x29
    XOR  {dst: reg}, {lhs: reg}, {rhs: reg} => 0x0D @ dst @ lhs @ rhs @ 0x2A
    
    ; In place
    ADD  {dst: reg}, {rhs: reg} => asm { ADD dst, dst, rhs }
    ADC  {dst: reg}, {rhs: reg} => asm { ADC dst, dst, rhs }
    SUB  {dst: reg}, {rhs: reg} => asm { SUB dst, dst, rhs }
    SBC  {dst: reg}, {rhs: reg} => asm { SBC dst, dst, rhs }
    AND  {dst: reg}, {rhs: reg} => asm { AND dst, dst, rhs }
    OR   {dst: reg}, {rhs: reg} => asm {  OR dst, dst, rhs }
    XOR  {dst: reg}, {rhs: reg} => asm { XOR dst, dst, rhs }
    
    ; Wide types
    ADD32 {dst: reg}, {lhs: reg}, {rhs: reg} => asm {
        ADD dst, lhs, rhs
        ADC dst+2, lhs+2, rhs+2
    }
    SUB32 {dst: reg}, {lhs: reg}, {rhs: reg} => asm {
        SUB dst, lhs, rhs
        SBC dst+2, lhs+2, rhs+2
    }
    ADD32_16 {dst: reg}, {lhs: reg}, {rhs: reg} => asm {
        ADD dst, lhs, rhs
        ADCI dst+2, lhs+2, 0
    }
    SUB32_16 {dst: reg}, {lhs: reg}, {rhs: reg} => asm {
        SUB dst, lhs, rhs
        SBCI dst+2, lhs+2, 0
    }
    ADD32 {dst: reg}, {rhs: reg} => asm { ADD32 dst, dst, rhs }
    SUB32 {dst: reg}, {rhs: reg} => asm { SUB32 dst, dst, rhs }
    ADD32_16 {dst: reg}, {rhs: reg} => asm { ADD32_16 dst, dst, rhs }
    SUB32_16 {dst: reg}, {rhs: reg} => asm { SUB32_16 dst, dst, rhs }
    
    ; Jumps & Branches
    JMP  {addr: u24} => 0x2B @ rvs24(addr)
    BRFS {addr: u24} => 0x2E @ rvs24(addr)
    BRFC {addr: u24} => 0x2F @ rvs24(addr)
    BREQ {val: reg}, {addr: u24} => 0x30 @ val @ rvs24(addr)
    BRNE {val: reg}, {addr: u24} => 0x31 @ val @ rvs24(addr)
    CALL {addr: u24} => 0x2C @ rvs24(addr)
    RET  => 0x2D
    
    ; Reg <- Value
    LDI  {dst: reg}, {val: i16} => 0x32 @ dst @ rvs16(val)
    LDIW {dst: reg}, {val: i32} => 0x33 @ dst @ rvs32(val)
    ; Aliases
    ;LDI  {dst: reg}, {val: i32} => asm { LDIW dst, val }
    
    ; Reg <- [Reg], [Reg] <- Reg
    LD   {dst: reg}, {ptr: reg} => 0x34 @ dst @ ptr
    LDB  {dst: reg}, {ptr: reg} => 0x35 @ dst @ ptr
    ST   {ptr: reg}, {src: reg} => 0x36 @ ptr @ src
    STB  {ptr: reg}, {src: reg} => 0x37 @ ptr @ src
    
    ; RAM2 access
    LD2W {ptr: reg} => 0x38 @ ptr
    LD2  {ptr: reg} => 0x39 @ ptr
    ST2W {ptr: reg} => 0x3A @ ptr
    ST2  {ptr: reg} => 0x3B @ ptr
    
    ; Reg <- Flash[Reg]
    LPM__ {dst: reg}, {ptr: reg} => 0x3C @ dst @ ptr
    LPB__ {dst: reg}, {ptr: reg} => 0x3D @ dst @ ptr
    
    ; OUT <- Reg
    OUT  {val: reg} => 0xBE @ val
    OUTI {val: i8} => asm {
        LDI TMP, val
        OUT TMP
    }
    
    ; OUT <- Reg with second CS line
    OUT2 {val: reg} => 0x7E @ val
    
    ; Special Nokia 5110 LCD instruction (OUT with both lines selected)
    OUT_LCD {val: reg} => 0xFE @ val
    
    ; Special instruction (AND) to output a single bit
    OUT_BIT {dst: reg}, {val: u8} =>
        0x0D @ 0x10 @ 0x0D @ dst @ dst @ 0x50 @ val @ 0x28
    OUT_BIT 1 => asm { OUT_BIT TMP, 0xC0 }
    OUT_BIT 0 => asm { OUT_BIT TMP, 0x00 }
    
    ; Constructed instructions:
    BREQ {lhs: reg}, {rhs: reg}, {addr: u24} => asm {
        SUB TMP, lhs, rhs
        BREQ TMP, addr
    }
    BREQI {lhs: reg}, {val: i16}, {addr: u24} => asm {
        SUBI TMP, lhs, val
        BREQ TMP, addr
    }
    BRNE {lhs: reg}, {rhs: reg}, {addr: u24} => asm {
        SUB TMP, lhs, rhs
        BRNE TMP, addr
    }
    BRNEI {lhs: reg}, {val: i16}, {addr: u24} => asm {
        SUBI TMP, lhs, val
        BRNE TMP, addr
    }
    BRLO {lhs: reg}, {rhs: reg}, {addr: u24} => asm {
        SUB TMP, lhs, rhs
        BRFS addr
    }
    BRLOI {lhs: reg}, {val: i16}, {addr: u24} => asm {
        SUBI TMP, lhs, val
        BRFS addr
    }
    BRSH {lhs: reg}, {rhs: reg}, {addr: u24} => asm {
        SUB TMP, lhs, rhs
        BRFC addr
    }
    BRSHI {lhs: reg}, {val: i16}, {addr: u24} => asm {
        SUBI TMP, lhs, val
        BRFC addr
    }

    ; Push / Pop
    PUSH_PC => asm {
        ST2W RAM_ADDR_SP
        ADDI RAM_ADDR_SP, 4
    }
    POP_PC => asm {
        SUBI RAM_ADDR_SP, 4
        LD2W RAM_ADDR_SP
    }
    PUSHW {val: reg} => asm {
        MOVW RAM_ADDR_SPVAL, val
        ST2W RAM_ADDR_SP
        ADDI RAM_ADDR_SP, 4
    }
    POPW {val: reg} => asm {
        SUBI RAM_ADDR_SP, 4
        LD2W RAM_ADDR_SP
        MOVW val, RAM_ADDR_SPVAL
    }
    PUSH {val: reg} => asm {
        MOV RAM_ADDR_SPVAL, val
        ST2 RAM_ADDR_SP
        ADDI RAM_ADDR_SP, 2
    }
    POP {val: reg} => asm {
        SUBI RAM_ADDR_SP, 2
        LD2 RAM_ADDR_SP
        MOV val, RAM_ADDR_SPVAL
    }

    ; LPM / LPB
    LPM {dst: reg}, {ptr: reg} => asm {
        REV32 TMP, ptr
        LPM__ dst, TMP + 1
    }
    LPB {dst: reg}, {ptr: reg} => asm {
        REV32 TMP, ptr
        LPB__ dst, TMP + 1
    }
}

; Start program
JMP Main

