#include <avr/io.h>
#include <util/delay.h>

//FCPU=8800000UL

#define SIG_CS (1<<0)
#define SIG_EN (1<<1)
#define SIG_CK (1<<2)
#define SIG_DT (1<<4)
#define SIG_PC (1<<3)

#define CLK_DELAY_US 10

/*******************************************************************************
 * Commands:
 *  f   |   $       | enter flashing mode
 *  q   |   -       | exit flashing mode
 *  s   |   S1S2    | read status register 1 & 2
 *
 *  + = ok, X = error value, M = crc mismatch
 *  pS1S2CC         |   write status register 1 & 2 "p2C00D1"
 *  rAAAAAALLLLLLCC |   read memory of length L
 *  eAAAAAASSCC     |   erase memory of size = 4 / 32 / 64 kB
 *  wAAAAAACC       |   + 256 values are expected
 ******************************************************************************/

/*******************************************************************************
 * UART routines
 ******************************************************************************/

static uint8_t receive_error;
static uint8_t buffer[32];

extern void asm_write_char(uint8_t c);
extern uint8_t asm_read_char(void);

void uart_start(void)
{
    _delay_us(100);
    PORTB |= SIG_PC;
    DDRB |= SIG_PC;
}

void uart_stop(void)
{
    DDRB &= ~SIG_PC;
    PORTB &= ~SIG_PC;
}

void uart_write_str4(uint8_t data)
{
    asm_write_char( data >= 10 ? 'A'+data-10 : '0'+data );
}

void uart_write_str(uint8_t data)
{
    uart_write_str4(data>>4);
    uart_write_str4(data&0xF);
}

uint8_t uart_read_str4(void)
{
    uint8_t c = asm_read_char();
    if (c >= '0' && c <= '9') {
        return c-'0';
    } else if (c >= 'A' && c <= 'F') {
        return c-'A'+10;
    } else if (c >= 'a' && c <= 'f') {
        return c-'a'+10;
    }
    receive_error = 1;
    return 0;
}

uint8_t uart_read_str(void)
{
    uint8_t data = uart_read_str4() << 4;
    data |= uart_read_str4();
    return data;
}

/*******************************************************************************
 * SPI routines
 ******************************************************************************/

void spi_start(void)
{
    PORTB &= ~SIG_CS;
    _delay_us(CLK_DELAY_US);
}

void spi_write(uint8_t data)
{
    for (uint8_t i = 0; i < 8; ++i) {
        PORTB &= ~SIG_CK;
        if (data & 0x80)
            PORTB |= SIG_DT;
        else
            PORTB &= ~SIG_DT;
        _delay_us(CLK_DELAY_US);
        data <<= 1;
        PORTB |= SIG_CK;
        _delay_us(CLK_DELAY_US);
    }
}

void spi_receive(void)
{
    DDRB &= ~SIG_DT;
    PORTB &= ~SIG_DT;
    _delay_us(CLK_DELAY_US);
}

uint8_t spi_read(void)
{
    uint8_t data = 0;
    for (uint8_t i = 0; i < 8; ++i) {
        data <<= 1;
        PORTB &= ~SIG_CK;
        _delay_us(CLK_DELAY_US);
        if (PINB & SIG_DT)
            data |= 1;
        PORTB |= SIG_CK;
        _delay_us(CLK_DELAY_US);
    }
    return data;
}

void spi_stop(void)
{
    PORTB &= ~SIG_CK;
    PORTB &= ~SIG_DT;
    _delay_us(CLK_DELAY_US);
    PORTB |= SIG_CS;
    _delay_us(CLK_DELAY_US);
    DDRB |= SIG_DT;
}

void spi_write_enable(void)
{
    spi_start();
    spi_write(0x06);
    spi_stop();
}

void spi_wait_for_busy(void)
{
    uint8_t x;
    do {
        spi_start();
        spi_write(0x05);
        spi_receive();
        x = spi_read();
        spi_stop();
    } while (x & 1);
}

/*******************************************************************************
 * Main
 ******************************************************************************/

int main (void)
{
    _delay_ms(100);
    
    /*
    DDRB |= SIG_EN;
    OCR0A = 0;
    TCCR0A = (1<<COM0B0) | (1<<WGM01);
    TCCR0B = 1;
    while(1);
    */
    
    PORTB = 0;
    DDRB = SIG_EN;
    
    // Outer idle loop
    while(1) {
        if (asm_read_char() == 0xFF) {
            // Start Flashing
            
            // Setup GPIO
            PORTB |= SIG_CS;
            PORTB |= SIG_EN;
            _delay_ms(10);
            DDRB |= SIG_CS;
            DDRB |= SIG_CK;
            DDRB |= SIG_DT;
            _delay_ms(20);
            
            uart_start();
            asm_write_char('\n');
            asm_write_char('$');
            uart_stop();
            
            while (1) {
                uint8_t cmd = asm_read_char();
                
                if (cmd == 0xFF) {
                    continue;
                } else if (cmd == 'q') {
                    // Quit flashing
                    break;
                } else if (cmd == 's') {
                    // Read status register
                    uart_start();
                    spi_start();
                    spi_write(0x05);
                    spi_receive();
                    uart_write_str(spi_read());
                    spi_stop();
                    spi_start();
                    spi_write(0x35);
                    spi_receive();
                    uart_write_str(spi_read());
                    spi_stop();
                    
                } else if (cmd == 'p') {
                    // Default protect first 256kB 0~0x03FFFF
                    // CMD: p2C00D1 to lock
                    // CMD: p0000A5 to unlock
                    receive_error = 0;
                    uint8_t sr1 = uart_read_str();
                    uint8_t sr2 = uart_read_str();
                    uint8_t crc = 0xA5 + sr1 + sr2;
                    crc -= uart_read_str();
                    uart_start();
                    if (receive_error) {
                        asm_write_char('X');
                    } else if (crc != 0) {
                        asm_write_char('M');
                    } else {
                        spi_write_enable();
                        spi_start();
                        spi_write(0x01);
                        spi_write(sr1);
                        spi_write(sr2);
                        spi_stop();
                        asm_write_char('+');
                    }
                    
                } else if (cmd == 'r' || cmd == 'e' || cmd == 'w') {
                    // Read / Erase / Write address
                    receive_error = 0;
                    uint8_t crc = 0xA5;
                    uint8_t addr1 = uart_read_str();
                    uint8_t addr2 = uart_read_str();
                    uint8_t addr3 = uart_read_str();
                    crc += addr1 + addr2 + addr3;
                    
                    if (cmd == 'r') {
                        // Read
                        uint8_t len1 = uart_read_str();
                        uint8_t len2 = uart_read_str();
                        uint8_t len3 = uart_read_str();
                        crc += len1 + len2 + len3;
                        crc -= uart_read_str();
                        uart_start();
                        if (receive_error) {
                            asm_write_char('X');
                        } else if (crc != 0) {
                            asm_write_char('M');
                        } else {
                            spi_start();
                            spi_write(0x03);
                            spi_write(addr1);
                            spi_write(addr2);
                            spi_write(addr3);
                            spi_receive();
                            uint32_t len = len3;
                            len |= (uint32_t)len2<<8;
                            len |= (uint32_t)len1<<16;
                            while(len--)
                                asm_write_char(spi_read());
                            spi_stop();
                            uart_stop();
                            continue;
                        }
                        
                    } else if (cmd == 'e') {
                        // Erase
                        uint8_t size = uart_read_str();
                        crc += size;
                        crc -= uart_read_str();
                        uart_start();
                        if (receive_error) {
                            asm_write_char('X');
                        } else if (crc != 0) {
                            asm_write_char('M');
                        } else if (size != 4 && size != 32 && size != 64) {
                            asm_write_char('Z');
                        } else {
                            spi_write_enable();
                            spi_start();
                            if (size == 4) {
                                spi_write(0x20);
                            } else if (size == 32) {
                                spi_write(0x52);
                            } else {
                                spi_write(0xD8);
                            }
                            spi_write(addr1);
                            spi_write(addr2);
                            spi_write(addr3);
                            spi_stop();
                            spi_wait_for_busy();
                            asm_write_char('+');
                        }
                    } else {
                        // Write
                        crc -= uart_read_str();
                        if (receive_error) {
                            uart_start();
                            asm_write_char('X');
                        } else if (crc != 0) {
                            uart_start();
                            asm_write_char('M');
                        } else {
                            spi_write_enable();
                            spi_start();
                            spi_write(0x02);
                            spi_write(addr1);
                            spi_write(addr2);
                            spi_write(addr3);
                            
                            for (uint8_t j = 0; j < 8; ++j) {
                                uart_start();
                                asm_write_char('+');
                                uart_stop();
                                for (uint8_t i = 0; i < 32; ++i)
                                    buffer[i] = asm_read_char();
                                for (uint8_t i = 0; i < 32; ++i)
                                    spi_write(buffer[i]);
                            }
                            spi_stop();
                            uart_start();
                            spi_wait_for_busy();
                            asm_write_char('+');
                        }
                    }
                } else {
                    // Command unknown
                    uart_start();
                }
                
                // Flashing mode prompt
                asm_write_char('\n');
                asm_write_char('$');
                uart_stop();
            }
            
            // Quit flashing
            DDRB = SIG_EN;
            _delay_ms(10);
            PORTB = 0;
        }
    }
    
    return (0);
}

