import serial
import glob
import time
import sys
import os

# Must be divisible by 64k
# so blocks can be erased
ADDR = 0x040000

BAUD = 19200
DEVICE = glob.glob('/dev/tty.wchusb*')
if len(DEVICE) == 0:
    print("No serial device available")
    exit(0)
DEVICE = DEVICE[0]

if len(sys.argv) < 2:
    print("No input file")
    exit(0)

if len(sys.argv) >= 3:
    ADDR = int(sys.argv[2], 16)

def get_crc(arr):
    crc = 0xA5
    for x in arr:
        while x:
            crc += x & 0xFF
            x >>= 8
    return crc & 0xFF

# Open file and serial port
f = open(sys.argv[1], 'rb')
ser = serial.Serial(DEVICE, BAUD, timeout=0.05)
ser.reset_input_buffer()

def stop():
    # Finish
    print("Finished")
    ser.close()
    f.close()
    exit(0)

def serial_read(size, timeout):
    t = time.time() + timeout
    data = b''
    while time.time() < t and len(data) < size:
        data += ser.read(size-len(data))
    return data

# Enter flashing mode
print("Entering flash mode...")
ser.write(b'\xFF')
msg = serial_read(2, 0.2)
if msg != b'\x0A$':
    failures = 0
    while True:
        ser.write(b'X')
        msg = serial_read(3, 0.2)
        if msg == b'X\x0A$':
            break
        failures += 1
        if failures > 8:
            print("Failed entering flashing mode")
            print("received:", msg)
            stop()
        print("Flushing buffer {}/8".format(failures))
        ser.write(b'\xFF'*32)
        serial_read(32+3, 0.5)
time.sleep(0.2)
ser.reset_input_buffer()

# Erase required space
f.seek(0, os.SEEK_END)
fsize = f.tell() & 0xFFFFFF
blocks_64k = (fsize + 0x007FFF) // 0x010000
fleft = max(0, fsize - blocks_64k *0x010000)
blocks_32k = (fleft + 0x003FFF) // 0x008000
fleft = max(0, fleft - blocks_32k *0x008000)
blocks_4k = (fleft + 0x000FFF) //  0x001000
addr = ADDR
blocks = ([(0x010000, 64)] * blocks_64k +
          [(0x008000, 32)] * blocks_32k +
          [(0x001000,  4)] * blocks_4k)
for b in blocks:
    print("Erasing {}k at 0x{:06X}~0x{:06X}...".format(
            b[1], addr, addr+b[0]))
    packet = "e{:06X}{:02X}{:02X}".format(
            addr, b[1], get_crc([addr, b[1]])).encode('ascii')
    ser.write(packet)
    msg = serial_read(14, 45)
    if msg != packet + b'+\x0A$':
        print("Failed, received:", msg)
        stop()
    addr += b[0]

# Write the file
f.seek(0)
addr = ADDR
print("Writing...")
for i in range(0, fsize, 256):
    packet = "w{:06X}{:02X}".format(addr, get_crc([addr])).encode('ascii')
    ser.write(packet)
    msg = serial_read(10, 0.5)
    if msg != packet + b'+':
        print("Failed writing bytes {}~{} at address 0x{:06X}".format(
                i, i+256, addr))
        print("received:", msg)
        stop()
    file_data = f.read(256)
    file_data += b'\xFF' * (256 - len(file_data))
    for j in range(0, 256, 32):
        ser.write(file_data[j:j+32])
        msg = serial_read(33, 0.5)
        if msg != file_data[j:j+32] + b'+':
            print("Failed writing bytes[] {}~{} at address 0x{:06X}".format(
                    i+j, i+j+32, addr))
            print("expected:", file_data[j:j+32] + b'+')
            print("received:", msg)
            stop()
    if serial_read(2, 0.5) != b'\n$':
        print("Failed writing bytes! {}~{} at address 0x{:06X}".format(
                    i, i+256, addr))
        stop()
    addr += 256
    print("Written {}/{}".format(i+256, (fsize + 0xFF) & 0xFFFF00))

# Verify
f.seek(0)
addr = ADDR
print("Verifying...")
for i in range(0, fsize, 256):
    packet = "r{:06X}{:06X}{:02X}".format(
            addr, 256, get_crc([addr, 256])).encode('ascii')
    ser.write(packet)
    file_data = f.read(256)
    file_data += b'\xFF' * (256 - len(file_data))
    msg = serial_read(15 + 256, 0.5)
    addr += 256
    if msg == packet + file_data:
        print("Verified {}/{}".format(i+256, (fsize + 0xFF) & 0xFFFF00))
    else:
        print("Failed verifying bytes {}~{} at address 0x{:06X}".format(
                i, i+256, addr))
        print("expected:", packet + file_data)
        print("received:", msg)
        stop()

# Done
print("Successfully flashed", fsize, "bytes of code")

# Exit flashing mode
ser.write(b'q')
ser.flush()
stop()

