#ifndef EPROM_H
#define EPROM_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <array>

#define CNT_BITS 8
#define OP_BITS 6

struct CPUeprom
{
    struct Datafield
    {
        unsigned char cs_ram2_flash : 1;
        unsigned char cs_ram1 : 1;
        unsigned char flash : 1;
        unsigned char ram1 : 1;
        unsigned char ram2 : 1;
        unsigned char data : 1;
        unsigned char flag : 1;
        unsigned char rst : 1;
        std::string msg;
        bool configured;
        
        Datafield() : configured(false) {};
        operator unsigned char() const {
            return (cs_ram2_flash<<0) | (cs_ram1<<1) | (flash<<2) | (ram1<<3) |
                   (ram2<<4) | (data<<5) | (flag<<6) | (rst<<7);
        };
    };
    
    void write(const std::string& path) {
        std::ofstream f(path, std::ios::binary | std::ios::trunc);
        for (auto& d : data) {
            if (!d.configured)
                std::cout << "Encountered uninitialized address\n";
            f << (unsigned char)d;
        }
        f.close();
    }
    
    static constexpr size_t size = 1<<17;
    std::array<Datafield, size> data;
    bool error_occured = false;
};

struct CPUaddr
{
    struct Address {
        unsigned int flag : 1;
        unsigned int flash : 1;
        unsigned int ram1_ram2 : 1;
        unsigned int cnt : CNT_BITS;
        unsigned int op : OP_BITS;
        
        Address(size_t a) {
            cnt = a & 0xFF;
            flag = (a>>10)&1;
            flash = (a>>11)&1;
            ram1_ram2 = (a>>9)&1;
            op =  (((a>>8)&1)<<0) | (((a>>13)&1)<<1) | (((a>>14)&1)<<2) |
                 (((a>>12)&1)<<3) | (((a>>15)&1)<<4) | (((a>>16)&1)<<5);
        };
        operator size_t() const {
            return  (cnt&0xFF)   |   ((flag&1)<<10)  |
                ((flash&1)<<11)  | ((ram1_ram2&1)<<9)|
                (((op>>0)&1)<<8) | (((op>>1)&1)<<13) | (((op>>2)&1)<<14) |
                (((op>>3)&1)<<12)| (((op>>4)&1)<<15) | (((op>>5)&1)<<16);
        };
    } addr;
    CPUeprom& eprom;
    
    CPUaddr(CPUeprom& eprom) : addr(0), eprom(eprom) {};
    CPUaddr(CPUeprom& eprom, size_t a) : addr(a), eprom(eprom) {};
    
    operator size_t() const {
        return (size_t)addr;
    };
    const CPUeprom::Datafield& get() const {
        return eprom.data[(size_t)addr];
    }
    void set(const CPUeprom::Datafield& data) {
        eprom.data[(size_t)addr] = data;
    }
};

std::ostream& operator<<(std::ostream& os, const CPUaddr& addr) {
    os << "Address = " << (size_t)addr << "\n";
    os << "R1 + R2: " << (int)addr.addr.ram1_ram2 << "\n";
    os << "flash  : " << (int)addr.addr.flash << "\n";
    os << "flag   : " << (int)addr.addr.flag << "\n";
    os << "cnt: " << (int)addr.addr.cnt << "\n";
    os << "op : " << (int)addr.addr.op << "\n";
    const auto& data = addr.get();
    os << "Data = '" << data.msg << "', ";
    os << (data.configured ? "configured" : "undefined") << "\n";
    os << "cs R2F: " << (int)data.cs_ram2_flash << "\n";
    os << "cs R1 : " << (int)data.cs_ram1 << "\n";
    os << "flash : " << (int)data.flash << "\n";
    os << "ram1  : " << (int)data.ram1 << "\n";
    os << "ram2  : " << (int)data.ram2 << "\n";
    os << "data  : " << (int)data.data << "\n";
    os << "flag  : " << (int)data.flag << "\n";
    os << "rst   : " << (int)data.rst << "\n";
    return os;
}

#endif
