#ifndef HW_H
#define HW_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <array>

template<int spi_size, int spi_width>
class SPIDevice {
public:
    struct input {
        unsigned char clk : 1;
        unsigned char din : 1;
        unsigned char cs : 1;
    };
    struct output {
        unsigned char out_0 : 1;
        unsigned char out_1 : 1;
    };
    SPIDevice() : cnt(0), addr(0) {};
    const output step(const input& in) {
        output output = { 0, 0 };
        
        if (!in.cs) {
            if (prev_in.cs)
                cnt = addr = cmd = wdata = 0;
            if (in.clk && !prev_in.clk && prev_in.cs) {
                std::cout << name << ": Invalid transition CLK/CS\n";
            } else if (in.clk && !prev_in.clk && cnt < 8+spi_width) {
                if (in.din != prev_in.din)
                    std::cout << name << ": Invalid transition ADDR\n";
                addr = (addr << 1) | (in.din & 1);
                if (++cnt == 8+spi_width) {
                    cmd = addr >> spi_width;
                    if (cmd == 2 || cmd == 3) {
                        addr &= (1<<spi_width)-1;
                        if (addr >= spi_size) {
                            std::cout << name << ": Out of bounds (";
                            std::cout << addr << ")\n";
                            addr %= spi_size;
                        }
                    } else {
                        std::cout << name << ": Unknown instruction (";
                        std::cout << (addr>>spi_width) << ")\n";
                    }
                }
            } else if (cnt >= 8+spi_width) {
                if (cmd == 3) {
                    if (!in.clk && prev_in.clk)
                        ++cnt;
                    if (cnt >= 8+spi_width+1) {
                        unsigned int idx = (cnt)-(8+spi_width+1);
                        unsigned char b = data[addr + idx/8];
                        if ((b >> (7 - idx%8)) & 1)
                            output.out_1 = 1;
                        else
                            output.out_0 = 1;
                    }
                } else if (cmd == 2 && in.clk && !prev_in.clk) {
                    if (in.din != prev_in.din)
                        std::cout << name << ": Invalid transition DATA\n";
                    wdata = (wdata << 1) | (in.din & 1);
                    unsigned int idx = (++cnt)-(8+spi_width);
                    if (idx % 8 == 0) {
                        //std::cout << name << "[" << addr << "] = ";
                        //std::cout << wdata << "\n";
                        data[addr++] = wdata;
                        wdata = 0;
                    }
                }
                
            }
        }
        prev_in = in;
        return output;
    };
    unsigned char get(size_t addr) {
        return data[addr];
    };
protected:
    std::array<unsigned char, spi_size> data;
    unsigned int cnt, addr, cmd, wdata;
    std::string name;
    input prev_in;
};

// W25Q80BV
/*
 * Default is Software write protection
 * SRP1=SRP0=0
 *
 * Instructions
 * 06h - write enable (00000110)
 * C7/60h - erase
 * 02h - 24bit addr - page program 256 bytes (last 8bits 0)
 * care on rising edge
 * 03h - 24bit addr 23 22 .. 2 1 0 - read one or more bytes
 * 7 6 5 4 3 2 1 0 7 6 5 ...
 */
class FLASH : public SPIDevice<16<<16, 24> {
public:
    FLASH(const std::string& name) {
        this->name = name;
        data.fill(0xFF);
        for (size_t i = 0; i < 1<<16; ++i) {
            data[2*i] = i>>8;
            data[2*i+1] = i;
        }
        for (size_t i = 0; i < 1<<16; ++i) {
            size_t x = (i & 0xF0F0) >> 4 | (i & 0x0F0F) << 4;
            x = (x & 0xCCCC) >> 2 | (x & 0x3333) << 2;
            x = (x & 0xAAAA) >> 1 | (x & 0x5555) << 1;
            data[(2<<16) + 2*i] = x;
            data[(2<<16) + 2*i+1] = x>>8;
        }
    };
    void program(const std::vector<unsigned char>& v) {
        auto idx = data.begin() + (4<<16);
        for (unsigned char c : v)
            *idx++ = c;
    };
    int program(const std::string& path, size_t addr) {
        std::ifstream f(path, std::ios::binary);
        auto idx = data.begin() + addr;
        char c;
        while (f.get(c))
            *idx++ = c;
        return idx-data.begin()-addr;
    };
    int program(const std::string& path) {
        return program(path, (4<<16));
    };
    void write(const std::string& path) {
        std::ofstream f(path, std::ios::binary | std::ios::trunc);
        for (size_t i = 0; i < 1<<18; ++i) {
            f << (unsigned char)data[i];
        }
        f.close();
    }
};

// 23k512
class RAM : public SPIDevice<1<<16, 16> {
public:
    RAM(const std::string& name) {
        this->name = name;
        data.fill(0);
    };
};

class HC595 {
public:
    struct input {
        unsigned char clk : 1;
        unsigned char din : 1;
        unsigned char rst : 1;
        unsigned char en : 1;
    };
    HC595() : data(0), output(0), data_valid(false) {};
    const unsigned char step(const input& in) {
        if (!in.rst) {
            data_valid = true;
            data = 0;
        } else {
            if (in.clk && !prev_in.clk && !prev_in.rst) {
                data_valid = false;
            } else if (in.clk && !prev_in.clk && in.en && !prev_in.en) {
                std::cout << "595: Invalid transition CLK/EN\n";
            } else if (in.clk && !prev_in.clk) {
                if (in.din != prev_in.din)
                    std::cout << "595: Invalid transition DATA\n";
                data = (data << 1) | (in.din & 1);
            } else if (in.en && !prev_in.en) {
                if (!data_valid)
                    std::cout << "595: Invalid transition CLK/RST\n";
                output = data;
            }
        }
        prev_in = in;
        return output;
    };
private:
    unsigned char data, output;
    bool data_valid;
    input prev_in;
};

// Out
class OUT {
public:
    struct input {
        unsigned char clk : 1;
        unsigned char din : 1;
        unsigned char cs : 1;
    };
    static void print_char(unsigned char c) {
        std::cout << "0x";
        std::cout << (char)((c>>4) >= 10 ? (c>>4)-10+'A' : (c>>4)+'0');
        std::cout << (char)((c&15) >= 10 ? (c&15)-10+'A' : (c&15)+'0');
    }
    void step(const input& in) {
        if (in.cs) {
            if (!prev_in.cs)
                cnt = data = datarev = 0;
            if (in.clk && !prev_in.clk && !prev_in.cs) {
                std::cout << "OUT: Invalid transition CLK/CS";
            } else if (in.clk && !prev_in.clk && cnt < 8) {
                if (in.din != prev_in.din)
                    std::cout << "OUT: Invalid transition DATA\n";
                data |= (in.din & 1) << cnt;
                datarev = (datarev << 1) | (in.din & 1);
                if (++cnt == 8) {
                    std::cout << "OUT:  ";
                    print_char(data);
                    std::cout << "    ";// << (int)data << "    ";
                    if (data >= 33 && data <= 126)
                        std::cout << (char)data  << "    ";
                    else
                        std::cout << "~"  << "    ";
                    print_char(datarev);
                    std::cout << "    ";// << (int)datarev << "    ";
                    if (datarev >= 33 && datarev <= 126)
                        std::cout << (char)datarev  << "    ";
                    else
                        std::cout << "~"  << "    ";
                    std::cout << "\n";
                }
                if (cnt > 8) {
                    std::cout << "OUT: Invalid count bits\n";
                }
            }
        }
        prev_in = in;
    };
private:
    unsigned char cnt, data, datarev;
    input prev_in;
};

// PCD8544
class PCD8544 {
public:
    struct input {
        unsigned char clk : 1;
        unsigned char din : 1;
        unsigned char cs : 1;
        unsigned char dc : 1;
    };
    void init() {
        for (size_t i = 0; i < 48; ++i)
            std::cout << "\n";
        std::cout << "\x1B[48A";
        std::cout << "\x1B[5G";
        std::cout.flush();
        x = y = 0;
    }
    void step(const input& in) {
        if (in.cs) {
            if (!prev_in.cs)
                cnt = data = datarev = 0;
            if (in.clk && !prev_in.clk && !prev_in.cs) {
                std::cout << "LCD: Invalid transition CLK/CS";
            } else if (in.clk && !prev_in.clk && cnt < 8) {
                if (in.din != prev_in.din)
                    std::cout << "LCD: Invalid transition DATA\n";
                data |= (in.din & 1) << cnt;
                datarev = (datarev << 1) | (in.din & 1);
                if (++cnt == 8) {
                    if (!in.dc) {
                        if (datarev & 0x80) {
                            x = datarev & 0x7F;
                            std::cout << "\x1B[" << (int)(2*x+5) << "G";
                        } else if (datarev & 0x40) {
                            int new_y = datarev & 0x3F;
                            if (new_y > y) {
                                std::cout << "\x1B[" << (int)(8*(new_y-y)) << "B";
                            } else if (new_y < y) {
                                std::cout << "\x1B[" << (int)(8*(y-new_y)) << "A";
                            }
                            y = new_y;
                        }
                        std::cout.flush();
                    } else {
                        for (int i = 0; i < 8; ++i) {
                            if (datarev & (1<<i)) {
                                std::cout << "\x1B[107m  ";
                            } else {
                                std::cout << "\x1B[44m  ";
                            }
                            std::cout << "\x1B[1B\x1B[2D";
                        }
                        std::cout << "\x1B[8A\x1B[2C";
                        if (++x >= 84) {
                            std::cout << "\x1B[5G";
                            x = 0;
                            if (++y >= 6) {
                                std::cout << "\x1B[40A";
                                y = 0;
                            } else
                                std::cout << "\x1B[8B";
                        }
                        //std::cout.flush();
                    }
                }
                if (cnt > 8)
                    std::cout << "LCD: Invalid count bits\n";
            }
        }
        prev_in = in;
    };
private:
    unsigned char cnt, data, datarev, x, y;
    input prev_in;
};

#endif
