#ifndef INSTRUCTION_SET_H
#define INSTRUCTION_SET_H

#include <iostream>
#include <initializer_list>
#include <string>
#include <vector>
#include "eprom.h"
#include "hw.h"

// Define number of operations and maximum counter value
#define CPU_CNT_MAX         (1<<CNT_BITS)
#define CPU_OP_MAX          (1<<OP_BITS)

// Define CPU inputs
#define CPU_INPUT_OPTIONS   8
#define CPU_INPUT_R1_R2     (1<<0)
#define CPU_INPUT_FLASH     (1<<1)
#define CPU_INPUT_FLAG      (1<<2)

// Define CPU output signals
#define CPU_SIGNAL_DATA     (1<<0)
#define CPU_SIGNAL_FLAG     (1<<1)
// Define other CPU signals
#define CPU_SIGNAL_CS_R2_F  (1<<2)
#define CPU_SIGNAL_CS_R1    (1<<3)
#define CPU_SIGNAL_FLASH    (1<<4)
#define CPU_SIGNAL_RAM1     (1<<5)
#define CPU_SIGNAL_RAM2     (1<<6)
#define CPU_SIGNAL_RST      (1<<7)

// Define 74HC595 shift register signals
#define CPU_595_DATA        CPU_SIGNAL_DATA
#define CPU_595_CLK         CPU_SIGNAL_FLASH
#define CPU_595_RST         CPU_SIGNAL_RAM1

// Other definitions
#define CPU_CLK_RAMS        (CPU_SIGNAL_RAM1 | CPU_SIGNAL_RAM2)
#define CPU_CLK_RAM1        (CPU_SIGNAL_RAM1)
#define CPU_CLK_RAM2        (CPU_SIGNAL_RAM2)
#define CPU_CLK_FLASH       (CPU_SIGNAL_FLASH)
#define CPU_CLK_ALL         (CPU_CLK_RAMS | CPU_CLK_FLASH)
#define CPU_CS_NONE         (CPU_SIGNAL_CS_R2_F | CPU_SIGNAL_CS_R1)
#define CPU_CS_RAM1         (CPU_SIGNAL_CS_R2_F)
#define CPU_CS_RAM2_FLASH   (CPU_SIGNAL_CS_R1)

/******************************************************************************
 *
 *
 *
 ******************************************************************************/

struct Transition
{
    typedef std::array<unsigned char, CPU_INPUT_OPTIONS> Lookup;
    static constexpr Lookup RESET0          = {0, 0, 0, 0, 0, 0, 0, 0};
    static constexpr Lookup RESET1          = {1, 1, 1, 1, 1, 1, 1, 1};
    static constexpr Lookup SET_FLAG        = {2, 2, 2, 2, 2, 2, 2, 2};
    static constexpr Lookup SET_FLAG_RAM    = {0, 2, 0, 2, 0, 2, 0, 2};
    static constexpr Lookup SET_FLAG_FLASH  = {0, 0, 2, 2, 0, 0, 2, 2};
    static constexpr Lookup DATA_0          = {0, 0, 0, 0, 2, 2, 2, 2};
    static constexpr Lookup DATA_1          = {1, 1, 1, 1, 3, 3, 3, 3};
    static constexpr Lookup COPY_R1_R2      = {0, 1, 0, 1, 2, 3, 2, 3};
    static constexpr Lookup COPY_FLASH      = {0, 0, 1, 1, 2, 2, 3, 3};
    static constexpr Lookup COPY_FLAG       = {0, 0, 0, 0, 3, 3, 3, 3};
    
    static constexpr Lookup TSTZ_RAM_first  = {2, 0, 2, 0, 2, 0, 2, 0};
    static constexpr Lookup TSTZ_RAM        = {0, 0, 0, 0, 2, 0, 2, 0};
    
    static constexpr Lookup COM_RAM_first   = {1, 2, 1, 2, 1, 2, 1, 2};
    static constexpr Lookup COM_RAM         = {1, 0, 1, 0, 1, 2, 1, 2};
    
    static constexpr Lookup AND_FLASH_first = {2, 2, 2, 1, 2, 2, 2, 1};
    static constexpr Lookup AND_FLASH       = {0, 0, 0, 1, 2, 2, 2, 1};
    
    static constexpr Lookup OR_FLASH_first  = {2, 1, 1, 1, 2, 1, 1, 1};
    static constexpr Lookup OR_FLASH        = {0, 1, 1, 1, 2, 1, 1, 1};
    
    static constexpr Lookup XOR_FLASH_first = {2, 1, 1, 2, 2, 1, 1, 2};
    static constexpr Lookup XOR_FLASH       = {0, 1, 1, 0, 2, 1, 1, 2};
    
    static constexpr Lookup NEG_RAM_first   = {2, 1, 2, 1, 2, 1, 2, 1};
    static constexpr Lookup NEG_RAM_tmp     = {1, 0, 1, 0, 2, 3, 2, 3};
    static constexpr Lookup NEG_RAM         = {1, 0, 1, 0, 2, 1, 2, 1};
    
    static constexpr Lookup ADD_RAM_0       = {0, 1, 0, 1, 0, 1, 0, 1};
    static constexpr Lookup ADD_RAM_1       = {1, 2, 1, 2, 1, 2, 1, 2};
    static constexpr Lookup ADC_RAM_0_tmp   = {0, 1, 0, 1, 3, 2, 3, 2};
    static constexpr Lookup ADC_RAM_0       = {0, 1, 0, 1, 1, 2, 1, 2};
    static constexpr Lookup ADC_RAM_1_tmp   = {1, 0, 1, 0, 2, 3, 2, 3};
    static constexpr Lookup ADC_RAM_1       = {1, 2, 1, 2, 2, 3, 2, 3};
    
    static constexpr Lookup ADD_FLASH_first = {0, 1, 1, 2, 0, 1, 1, 2};
    static constexpr Lookup ADC_FLASH_tmp   = {0, 1, 1, 0, 3, 2, 2, 3};
    static constexpr Lookup ADC_FLASH       = {0, 1, 1, 2, 1, 2, 2, 3};
    
    static constexpr Lookup SUB_RAM_0       = {0, 1, 0, 1, 0, 1, 0, 1};
    static constexpr Lookup SUB_RAM_1       = {3, 0, 3, 0, 3, 0, 3, 0};
    static constexpr Lookup SBC_RAM_0_tmp   = {0, 1, 0, 1, 3, 2, 3, 2};
    static constexpr Lookup SBC_RAM_0       = {0, 1, 0, 1, 3, 0, 3, 0};
    static constexpr Lookup SBC_RAM_1_tmp   = {1, 0, 1, 0, 2, 3, 2, 3};
    static constexpr Lookup SBC_RAM_1       = {3, 0, 3, 0, 2, 3, 2, 3};
    
    static constexpr Lookup SUB_FLASH_first = {0, 1, 3, 0, 0, 1, 3, 0};
    static constexpr Lookup SBC_FLASH_tmp   = {0, 1, 1, 0, 3, 2, 2, 3};
    static constexpr Lookup SBC_FLASH       = {0, 1, 3, 0, 3, 0, 2, 3};
};

struct Mask
{
    static constexpr unsigned char MASK_ALL = 0xFF;
    static constexpr unsigned char MASK_FLAG0 = 0x0F;
    static constexpr unsigned char MASK_FLAG1 = 0xF0;
};

class InstructionSet
{
public:
    InstructionSet(CPUeprom& eprom) :
        current_lookup(&Transition::RESET0),
        current_mask(Mask::MASK_ALL), current_cmd(0),
        eprom(eprom), cnt_(0), cnt(0), op(0),
        err(false) {
        current_lookup_data[0] = &Transition::RESET0;
        current_lookup_data[1] = &Transition::RESET1;
    };
    
    size_t opcodeGet() const {
        return op;
    }
    
    void opcode(size_t new_op, const std::string& msg) {
        current_mask = Mask::MASK_ALL;
        current_cmd = 0;
        next_msg = msg;
        err = false;
        op = new_op;
        cnt_ = 0;
        cnt = 0;
    };
    
    void opcodeNext(const std::string& msg) {
        opcode(op + 1, msg);
    }
    
    void labelNext(const std::string& msg) {
        next_msg = msg;
    }
    
    void addCmd(unsigned char cmd, const Transition::Lookup& lookup) {
        current_lookup = &lookup;
        current_cmd = cmd;
        write();
    }
    
    void fill() {
        CPUaddr a(eprom);
        a.addr.op = op;
        if (!a.get().msg.empty()) {
            std::cout << "OP: ";
            OUT::print_char(op);
            std::cout << " (" << a.get().msg << ")";
            std::cout << "    ~ " << cnt_ << "\n";
        }
        while (cnt_ < CPU_CNT_MAX)
            addCmd(CPU_CS_NONE, *current_lookup_data[0]);
    }
    
    void repeat(size_t r) {
        for (size_t i = 0; i < r; ++i)
            write();
    }
    
    void setMask(unsigned char mask) {
        current_mask = mask;
        if (mask != Mask::MASK_ALL)
            cnt_ = cnt;
    }
    
    void setTransitions(const Transition::Lookup& lookup0,
                        const Transition::Lookup& lookup1) {
        current_lookup_data[0] = &lookup0;
        current_lookup_data[1] = &lookup1;
    };
    
    void clock(size_t bits, const Transition::Lookup& lookup,
               unsigned char idle, unsigned char clock) {
        for (size_t i = 0; i < bits; ++i) {
            addCmd(idle, lookup);
            addCmd(idle ^ clock, lookup);
        }
    }
    
    void clock(size_t bits,
               const Transition::Lookup& lookup_tmp,
               const Transition::Lookup& lookup,
               unsigned char idle, unsigned char clock) {
        for (size_t i = 0; i < bits; ++i) {
            addCmd(idle, lookup_tmp);
            addCmd(idle ^ clock, lookup);
        }
    }
    
    void clock(size_t bits, size_t data,
               unsigned char idle, unsigned char clock) {
        for (size_t i = 0; i < bits; ++i) {
            size_t l = ((1UL<<(bits-1)) & (data<<i)) ? 1 : 0;
            addCmd(idle, *current_lookup_data[l]);
            addCmd(idle ^ clock, *current_lookup_data[l]);
        }
    }
    
    void clockReverse(size_t bits, size_t data,
                      unsigned char idle, unsigned char clock) {
        for (size_t i = 0; i < bits; ++i) {
            size_t l = ((1UL<<i) & data) ? 1 : 0;
            addCmd(idle, *current_lookup_data[l]);
            addCmd(idle ^ clock, *current_lookup_data[l]);
        }
    }
    
    void addValue(size_t bits, size_t value,
                  unsigned char idle, unsigned char clk) {
        clock(1, (1 & value) ? Transition::ADD_RAM_1 :
                               Transition::ADD_RAM_0, idle, clk);
        const Transition::Lookup* t[2][2] =
            { {&Transition::ADC_RAM_0_tmp, &Transition::ADC_RAM_0},
              {&Transition::ADC_RAM_1_tmp, &Transition::ADC_RAM_1} };
        for(size_t i = 1; i < bits; ++i) {
            size_t l = ((1UL<<i) & value) ? 1 : 0;
            clock(1, *t[l][0], *t[l][1], idle, clk);
        }
    }
    
    void subValue(size_t bits, size_t value,
                  unsigned char idle, unsigned char clk) {
        clock(1, (1 & value) ? Transition::SUB_RAM_1 :
                               Transition::SUB_RAM_0, idle, clk);
        const Transition::Lookup* t[2][2] =
            { {&Transition::SBC_RAM_0_tmp, &Transition::SBC_RAM_0},
              {&Transition::SBC_RAM_1_tmp, &Transition::SBC_RAM_1} };
        for(size_t i = 1; i < bits; ++i) {
            size_t l = ((1UL<<i) & value) ? 1 : 0;
            clock(1, *t[l][0], *t[l][1], idle, clk);
        }
    }
    
    void jump(size_t op, unsigned char idle) {
        clock(OP_BITS, op, idle | CPU_595_RST, CPU_595_CLK);
        addCmd(idle | CPU_595_RST | CPU_SIGNAL_RST, *current_lookup_data[0]);
    }

private:
    const Transition::Lookup* current_lookup_data[2];
    const Transition::Lookup* current_lookup;
    unsigned char current_mask;
    unsigned char current_cmd;
    std::string next_msg;
    CPUeprom& eprom;
    size_t cnt_;
    size_t cnt;
    size_t op;
    bool err;
    
    void write() {
        if (err)
            return;
        if (op >= CPU_OP_MAX) {
            std::cout << "IS: opcode overflow";
            OUT::print_char(op);
            std::cout << ")\n";
            err = eprom.error_occured = true;
            return;
        }
        if (cnt_ >= CPU_CNT_MAX) {
            std::cout << "IS: cnt overflow, op = ";
            OUT::print_char(op);
            CPUaddr a(eprom); a.addr.op = op;
            if (!a.get().msg.empty())
                std::cout << " (" << a.get().msg << ")";
            std::cout << "\n";
            eprom.error_occured = true;
            return;
        }
        
        CPUaddr addr(eprom);
        addr.addr.cnt = cnt_++;
        addr.addr.op = op;
        if (current_mask == Mask::MASK_ALL)
            cnt = cnt_;
        
        bool overwriten = false;
        for (size_t in = 0; in < CPU_INPUT_OPTIONS; ++in) {
            if (!(current_mask & (1<<in)))
                continue;
        
            addr.addr.ram1_ram2 =   (in & CPU_INPUT_R1_R2)  ? 1 : 0;
            addr.addr.flash =       (in & CPU_INPUT_FLASH)  ? 1 : 0;
            addr.addr.flag =        (in & CPU_INPUT_FLAG)   ? 1 : 0;
            
            CPUeprom::Datafield output;
            unsigned char data = current_cmd | (*current_lookup)[in];
            output.cs_ram2_flash = (data & CPU_SIGNAL_CS_R2_F) ? 1 : 0;
            output.cs_ram1= (data & CPU_SIGNAL_CS_R1)   ? 1 : 0;
            output.flash =  (data & CPU_SIGNAL_FLASH)   ? 1 : 0;
            output.ram1 =   (data & CPU_SIGNAL_RAM1)    ? 1 : 0;
            output.ram2 =   (data & CPU_SIGNAL_RAM2)    ? 1 : 0;
            output.data =   (data & CPU_SIGNAL_DATA)    ? 1 : 0;
            output.flag =   (data & CPU_SIGNAL_FLAG)    ? 1 : 0;
            output.rst =    (data & CPU_SIGNAL_RST)     ? 1 : 0;
            
            output.configured = true;
            if (!next_msg.empty())
                output.msg = next_msg;
            if (addr.get().configured)
                overwriten = true;
            addr.set(output);
        }
        if (!next_msg.empty())
            next_msg.clear();
        if (overwriten) {
            std::cout << "IS: overwriting cmd\n";
            addr.eprom.error_occured = true;
        }
    };
};

/******************************************************************************
 *
 *
 *
 ******************************************************************************/

void check_eprom(CPUeprom& eprom) {
    CPUaddr addr(eprom, 0);
    for (size_t op = 0; op < CPU_OP_MAX; ++op) {
        addr.addr.op = op;
        int cnt_any_is_configured = 0;
        int cnt_any_is_undefined = 0;
        for (size_t cnt = 0; cnt < CPU_CNT_MAX; ++cnt) {
            addr.addr.cnt = cnt;
            bool any_is_configured = false;
            bool any_is_undefined = false;
            for (size_t in = 0; in < CPU_INPUT_OPTIONS; ++in) {
                addr.addr.ram1_ram2 =   (in & CPU_INPUT_R1_R2)  ? 1 : 0;
                addr.addr.flash =       (in & CPU_INPUT_FLASH)  ? 1 : 0;
                addr.addr.flag =        (in & CPU_INPUT_FLAG)   ? 1 : 0;
                bool cfg = addr.get().configured;
                any_is_configured |= cfg;
                any_is_undefined |= !cfg;
            }
            cnt_any_is_configured += any_is_configured;
            cnt_any_is_undefined += any_is_undefined;
            if (any_is_configured && any_is_undefined) {
                std::cout << "IS: op = ";
                OUT::print_char(op);
                std::cout << ", cnt = " << cnt;
                std::cout << " is not complete\n";
                eprom.error_occured = true;
                break;
            }
        }
        if (cnt_any_is_configured && cnt_any_is_undefined) {
            addr.addr.cnt = 0;
            std::cout << "WARNING IS: op = ";
            OUT::print_char(op);
            if (!addr.get().msg.empty())
                std::cout << " (" << addr.get().msg << ")";
            std::cout << ", " << cnt_any_is_configured << "/";
            std::cout << (cnt_any_is_configured+cnt_any_is_undefined);
            std::cout << "\n";
        }
        if (!cnt_any_is_configured) {
            std::cout << "WARNING IS: op = ";
            OUT::print_char(op);
            std::cout << " is missing\n";
        }
    }
}

#endif
