#ifndef INSTRUCTIONS_H
#define INSTRUCTIONS_H

#include <iostream>
#include <initializer_list>
#include <string>
#include <vector>
#include "instruction_set.h"
#include "eprom.h"

/******************************************************************************
 *
 *
 *
 ******************************************************************************/

// Static instructions:
#define OP_CODE_INIT        0
#define OP_CODE_RESET       1
#define OP_CODE_FETCH       2
#define OP_CODE_FETCH_      3
#define OP_CODE_INCR3       4
#define OP_CODE_INCR5       5
#define OP_CODE_INCR7       6
#define OP_CODE_INCR8       7
#define OP_CODE_CPY32       8
#define OP_CODE_CPY16       9
#define OP_CODE_CPYPC       10
#define OP_CODE_ST16        11
#define OP_CODE_ST8         12
#define OP_CODE_EXT         13
#define OP_CODE_NOP         16
#define OP_CODE_HALT        (CPU_OP_MAX-1)

// SPI Read/Write commands
#define SPI_WRITE       0x02
#define SPI_READ        0x03

// RAM1: 0:  -- , 1~3: PC (copy), 4~7: PC (rev), 8~9: SP, 10~13: SPV
// RAM2: 0: Flag, 1~3: PC (main), 4~5: PTR,   6~9: VAL  , 10~: stack data
#define RAM_ADDR_FLAG   0
#define RAM_ADDR_PC     1
#define RAM_ADDR_PTR    4
#define RAM_ADDR_VAL    6
#define RAM_ADDR_SP     8
#define RAM_ADDR_SPV    10

#define FLASH_IDENTITY  0x000000
#define FLASH_REVERSE   0x020000
#define INIT_PC         0x040000
#define INIT_SP         0x000A

/******************************************************************************
 *
 *
 *
 ******************************************************************************/

void instruction_set_fill(CPUeprom& eprom)
{
    InstructionSet is(eprom);
    is.setTransitions(Transition::RESET0, Transition::RESET1);
    
    auto setReadWrite = [&](unsigned char r, unsigned char w) mutable {
        if (r | w)
            is.clock(7, 1, 0, r | w);
        if (r)
            is.clock(1, SPI_READ, 0, r);
        if (w)
            is.clock(1, SPI_WRITE, 0, w);
    };
    
    auto goFlashReverse = [&](unsigned char clk) mutable {
        is.clock(6, 0, 0, CPU_CLK_FLASH | clk);
        is.clock(1, 1, 0, CPU_CLK_FLASH);
    };
    
    auto copy_R1_to = [&](unsigned char clk, size_t w) mutable {
        is.clock(w, Transition::COPY_R1_R2, CPU_CLK_RAM1, CPU_CLK_RAM1 | clk);
    };
    auto copy_R2_to = [&](unsigned char clk, size_t w) mutable {
        is.clock(w, Transition::COPY_R1_R2, CPU_CLK_RAM2, CPU_CLK_RAM2 | clk);
    };
    auto copy_Flash_to = [&](unsigned char clk, size_t w) mutable {
        is.clock(w, Transition::COPY_FLASH, CPU_CLK_FLASH, CPU_CLK_FLASH | clk);
    };
    
    auto computeR1 = [&](size_t w, const Transition::Lookup& l) mutable {
        is.clock(w, l, CPU_CLK_RAM1, CPU_CLK_RAMS);
    };
    auto computeR1_ = [&](size_t w,
                          const Transition::Lookup& l_tmp,
                          const Transition::Lookup& l) mutable {
        is.clock(w, l_tmp, l, CPU_CLK_RAM1, CPU_CLK_RAMS);
    };
    auto computeFlash = [&](size_t w, const Transition::Lookup& l) mutable {
        is.clock(w, l, CPU_CLK_RAM1 | CPU_CLK_FLASH, CPU_CLK_ALL);
    };
    auto computeFlash_ = [&](size_t w,
                             const Transition::Lookup& l_tmp,
                             const Transition::Lookup& l) mutable {
        is.clock(w, l_tmp, l, CPU_CLK_RAM1|CPU_CLK_FLASH, CPU_CLK_ALL);
    };
    auto computeR2 = [&](size_t w, const Transition::Lookup& l) mutable {
        is.clock(w, l, CPU_CLK_RAM2 | CPU_CLK_FLASH, CPU_CLK_ALL);
    };
    auto computeR2_ = [&](size_t w,
                          const Transition::Lookup& l_tmp,
                          const Transition::Lookup& l) mutable {
        is.clock(w, l_tmp, l, CPU_CLK_RAM2 | CPU_CLK_FLASH, CPU_CLK_ALL);
    };
    
    auto copy_result = [&](size_t w) mutable {
        setReadWrite(CPU_CLK_RAM2, CPU_CLK_RAM1); // R2 read, R1 write
        is.clock(16, RAM_ADDR_PTR, 0, CPU_CLK_RAM2); // PTR -> R2[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R2_to(CPU_CLK_RAM1, 16); // R2[PTR] -> R1[]
        copy_R2_to(CPU_CLK_RAM1, w); // R2[VAL] -> R1[addr]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
    };
    
    auto copy_pc = [&]() mutable {
        setReadWrite(CPU_CLK_RAM1, CPU_CLK_RAM2); // R1 read, R2 write
        is.clock(16, RAM_ADDR_PC, 0, CPU_CLK_RAMS); // PC -> R1[] & R2[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R1_to(CPU_CLK_RAM2, 24); // R1[PC] -> R2[PC]
        is.addCmd(0, Transition::DATA_0); // Save write
    };
    
    auto incr_pure = [&](size_t value, bool flag) mutable {
        setReadWrite(CPU_CLK_RAM1, CPU_CLK_RAM2); // R1 read, R2 write
        if (flag) {
            is.clock(16, RAM_ADDR_FLAG, 0, CPU_CLK_RAMS); // FLAG -> R1[] & R2[]
            is.addCmd(0, Transition::DATA_0); // Enable read
            // R1[FLAG] -> R2[FLAG]
            is.clock(7, Transition::DATA_0, CPU_CLK_RAM1, CPU_CLK_RAMS);
            is.clock(1, Transition::COPY_FLAG, CPU_CLK_RAM1, CPU_CLK_RAMS);
        } else {
            is.clock(16, RAM_ADDR_PC, 0, CPU_CLK_RAMS); // PC -> R1[] & R2[]
            is.addCmd(0, Transition::DATA_0); // Enable read
        }
        is.setMask(Mask::MASK_ALL);
        is.addValue(24, value, CPU_CLK_RAM1, CPU_CLK_RAMS); // R1[PC] -> R2[PC]
        is.addCmd(0, Transition::DATA_0); // Save write
    };
    
/******************************************************************************
 *
 *
 *
 ******************************************************************************/

    // INIT: Set RAM status byte to sequential read/write
    is.opcode(OP_CODE_INIT, "INIT");
    is.addCmd(CPU_CS_NONE, Transition::RESET0);
    is.repeat(CPU_CNT_MAX-1-32-2-13);
    is.clock(8, 0x01, 0, CPU_CLK_RAMS); // Write status byte
    is.clock(8, 0x40, 0, CPU_CLK_RAMS); // Set sequential mode
    is.addCmd(0, Transition::RESET0); // Save write
    is.addCmd(CPU_CS_NONE, Transition::RESET0); // Stop
    is.jump(OP_CODE_RESET, CPU_CS_NONE);
    is.fill();
    
    // RESET: reset program counter and stack pointer
    is.opcode(OP_CODE_RESET, "RESET");
    is.clock(8, SPI_WRITE, 0, CPU_CLK_RAMS);
    is.clock(10, 0, 0, CPU_CLK_RAMS);
    // Program counter:
    is.clock(6, RAM_ADDR_FLAG, 0, CPU_CLK_RAM2); // FLAG -> RAM2
    is.clock(8, 0, 0, CPU_CLK_RAM2); // Flag
    is.clockReverse(24, INIT_PC, 0, CPU_CLK_RAM2);
    // Stack pointer
    is.clock(6, RAM_ADDR_SP, 0, CPU_CLK_RAM1); // SP -> RAM1
    is.clockReverse(16, INIT_SP, 0, CPU_CLK_RAM1);
    // Zero other things
    is.clock(32, 0, 0, CPU_CLK_RAMS);
    is.addCmd(0, Transition::RESET0);  // Save write
    is.addCmd(CPU_CS_NONE, Transition::RESET0); // Stop
    is.jump(OP_CODE_FETCH, CPU_CS_NONE);
    is.fill();
    
    // HALT:
    is.opcode(OP_CODE_HALT, "HALT");
    is.addCmd(CPU_CS_NONE, Transition::RESET0); // Reset 595
    is.jump(OP_CODE_HALT, CPU_CS_NONE);
    is.fill();
    
    // Flag oriented instructions
    is.setTransitions(Transition::DATA_0, Transition::DATA_1);
    
    // FETCH:
    is.opcode(OP_CODE_FETCH, "_FETCH");
    setReadWrite(CPU_CLK_RAM2 | CPU_CLK_FLASH, CPU_CLK_RAM1);
    goFlashReverse(CPU_CLK_RAMS); // Flash reverse
    is.clock(6, 0, 0, CPU_CLK_ALL); // Flash reverse + 0x00
    is.clock(2, 0, 0, CPU_CLK_FLASH); // + 0x00
    is.clock(4, RAM_ADDR_FLAG, 0, CPU_CLK_RAMS); // FLAG -> R1[] & R2[]
    // Inputs
    is.clock(1, Transition::COPY_FLASH, CPU_CLK_RAM2, CPU_CLK_RAM1);
    is.clock(1, Transition::COPY_R1_R2, CPU_CLK_RAM2, CPU_CLK_RAM1);
    is.clock(2, Transition::DATA_0, 0, CPU_CLK_RAM2);
    is.clock(5, Transition::DATA_0, 0, CPU_CLK_RAMS);
    is.addCmd(0, Transition::DATA_0); // Enable read R2 7th bit
    // Set FLAG
    is.addCmd(0, Transition::SET_FLAG_RAM);
    is.clock(1, Transition::COPY_FLAG, CPU_CLK_RAM2, CPU_CLK_RAMS);
    copy_R2_to(CPU_CLK_RAM1, 16); // R2[PC/0/1] -> R1[PC/0/1]
    copy_R2_to(CPU_CLK_RAM1 | CPU_CLK_FLASH, 8); // R2[PC/2]->R1[PC/2] & Flash
    is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash reverse enable
    is.addCmd(0, Transition::DATA_0); // Enable read
    copy_Flash_to(CPU_CLK_RAM1, 8); // Flash[rev] -> R1[PTR]
    // ---
    is.jump(OP_CODE_FETCH_, CPU_CS_RAM1);
    is.fill();
    is.opcode(OP_CODE_FETCH_, "_FETCH_");
    setReadWrite(CPU_CLK_RAM2 | CPU_CLK_FLASH, 0); // R2/Flash read
    //goFlashReverse(CPU_CLK_RAM2); // Flash reverse
    // Saves 1 clock to inline it (No need for enable)
    is.clock(6, 0, 0, CPU_CLK_FLASH | CPU_CLK_RAM2);
    is.clock(10, RAM_ADDR_PC, 0, CPU_CLK_RAM2); // PC -> R2[]
    is.clock(1, 1, 0, CPU_CLK_FLASH);
    //is.addCmd(0, Transition::DATA_0); // Enable read
    copy_R2_to(CPU_CLK_FLASH, 16); // R2[PC/0/1] -> Flash[]
    is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash reverse enable
    is.addCmd(0, Transition::DATA_0); // Enable read
    copy_Flash_to(CPU_CLK_RAM1, 16); // Flash[rev] -> R1[PTR+]
    is.addCmd(0, Transition::DATA_0); // Save write
    is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
    setReadWrite(CPU_CLK_RAM1 | CPU_CLK_FLASH, CPU_CLK_RAM2);
    is.clock(16, RAM_ADDR_PTR, 0, CPU_CLK_RAMS); // PTR -> R1[] & R2[]
    is.addCmd(0, Transition::DATA_0); // Enable read
    copy_R1_to(CPU_CLK_FLASH, 23); // R1[PTR] -> Flash
    // Combine Last clock cycle and Stop R1
    is.addCmd(CPU_CLK_RAM1, Transition::COPY_R1_R2);
    is.addCmd(CPU_CLK_FLASH | CPU_CS_RAM2_FLASH, Transition::COPY_R1_R2);
    is.clock(7, 1, 0, CPU_CLK_RAM1); // 0000001 -> R1
    // Flash -> 595, Special care must be taken
    for (size_t i = 0; i < 8; ++i) {
        is.addCmd(CPU_595_RST, Transition::COPY_FLASH); // Sets data line
        is.addCmd(CPU_CLK_FLASH|CPU_595_RST, Transition::COPY_FLASH); // Tick
        if (i < 7) // The last bit is not necessary
            is.addCmd(CPU_595_RST, Transition::COPY_FLASH); // Tick low
    }
    is.addCmd(CPU_595_RST | CPU_SIGNAL_RST, Transition::DATA_1);
    is.fill();
    
    // INCRn:
    auto incr = [&](size_t op, const std::string& name, size_t value) mutable {
        is.opcode(op, name);
        incr_pure(value, true);
        is.jump(OP_CODE_FETCH, CPU_CS_NONE);
        is.fill();
    };
    incr(OP_CODE_INCR3, "_INCR3", 3);
    incr(OP_CODE_INCR5, "_INCR5", 5);
    incr(OP_CODE_INCR7, "_INCR7", 7);
    incr(OP_CODE_INCR8, "_INCR8", 8);
    
    // COPY:
    auto cpy = [&](size_t op, const std::string& name, size_t w) mutable {
        is.opcode(op, name);
        copy_result(w);
        is.jump(OP_CODE_INCR5, CPU_CS_NONE);
        is.fill();
    };
    cpy(OP_CODE_CPY32, "_CPY32", 32);
    cpy(OP_CODE_CPY16, "_CPY16", 16);
    
    // CALLCOPY:
    is.opcode(OP_CODE_CPYPC, "_CALL_");
    copy_pc();
    is.jump(OP_CODE_FETCH, CPU_CS_NONE);
    is.fill();
    
    // STCOPY:
    auto stcpy = [&](size_t op, const std::string& name, size_t w) mutable {
        is.opcode(op, name);
        setReadWrite(CPU_CLK_RAM2 | CPU_CLK_FLASH, CPU_CLK_RAM1);
        goFlashReverse(CPU_CLK_RAM2); // Flash reverse
        is.clock(10, RAM_ADDR_PTR, 0, CPU_CLK_RAM2); // PTR -> R2[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R2_to(CPU_CLK_FLASH, 16); // R2[PTR] -> Flash[]
        is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash reverse enable
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        copy_R2_to(CPU_CLK_RAM1, w); // R2[VAL] -> R1[addr]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_INCR5, CPU_CS_NONE);
        is.fill();
    };
    stcpy(OP_CODE_ST16, "_ST16", 16);
    stcpy(OP_CODE_ST8, "_ST8", 8);
    
    // EXT: Parameters extension
    is.opcode(OP_CODE_EXT, "_EXT");
    is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
    copy_Flash_to(CPU_CLK_RAM2, 16); // Flash -> R2[PTR]
    copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
    is.addCmd(0, Transition::DATA_0); // Enable read
    copy_R1_to(CPU_CLK_RAM2, 16); // R1[addr] -> R2[VAL]
    is.addCmd(CPU_CS_RAM2_FLASH, Transition::DATA_0); // Stop
    setReadWrite(CPU_CLK_RAM1, 0); // R1 read
    copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
    // Flash -> 595, Special care must be taken
    for (size_t i = 0; i < 8; ++i) {
        is.addCmd(CPU_595_RST, Transition::COPY_FLASH); // Sets data line
        is.addCmd(CPU_CLK_FLASH|CPU_595_RST, Transition::COPY_FLASH); // Tick
        is.addCmd(CPU_595_RST, Transition::COPY_FLASH); // Tick low
    }
    is.addCmd(CPU_595_RST | CPU_SIGNAL_RST, Transition::DATA_1);
    is.fill();
    
    // STF|CLF|NOP:
    auto singleton = [&](const std::string& name,
                         const Transition::Lookup& t) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        is.addCmd(CPU_CS_NONE, t);
        incr_pure(1, true);
        is.jump(OP_CODE_FETCH, CPU_CS_NONE);
        is.fill();
    };
    singleton("STF", Transition::SET_FLAG);
    singleton("CLF", Transition::RESET0);
    singleton("NOP", Transition::DATA_0);
    
/******************************************************************************
 *
 *
 *
 ******************************************************************************/

    // Helper function
    auto addOp = [&](const std::string& name, unsigned char incrN,
                     std::function<void()> f) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM2, 16); // Flash -> R2[PTR]
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        f(); // R1[addr] -> R2[VAL]
        // Copy R2 -> R1
        if (incrN) {
            is.addCmd(0, Transition::DATA_0); // Save write
            is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
            copy_result(16);
            is.jump(incrN, CPU_CS_NONE);
        }
        is.fill();
    };
    
    // MOV|INC|DEC|COM|NEG|LSL|LSR|ROL|ROR|ASR|REV:
    addOp("MOV", OP_CODE_INCR5, [&] {
        //computeR1(16, Transition::COPY_R1_R2);
        copy_R1_to(CPU_CLK_RAM2, 16);
    });
    addOp("MOV32", 0, [&] {
        //computeR1(32, Transition::COPY_R1_R2);
        copy_R1_to(CPU_CLK_RAM2, 32);
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_CPY32, CPU_CS_NONE);
    });
    addOp("INC", OP_CODE_INCR5, [&] {
        is.addValue(16, 1, CPU_CLK_RAM1, CPU_CLK_RAMS);
    });
    addOp("DEC", OP_CODE_INCR5, [&] {
        is.subValue(16, 1, CPU_CLK_RAM1, CPU_CLK_RAMS);
    });
    addOp("COM", OP_CODE_INCR5, [&] {
        computeR1(1, Transition::COM_RAM_first);
        computeR1(15, Transition::COM_RAM);
    });
    addOp("NEG", OP_CODE_INCR5, [&] {
        computeR1(1, Transition::NEG_RAM_first);
        computeR1_(15, Transition::NEG_RAM_tmp, Transition::NEG_RAM);
    });
    addOp("LSL", OP_CODE_INCR5, [&] {
        is.clock(1, Transition::DATA_0, 0, CPU_CLK_RAM2);
        is.clock(15, Transition::COPY_R1_R2, CPU_CLK_RAM1, CPU_CLK_RAMS);
        is.clock(1, Transition::SET_FLAG_RAM, CPU_CLK_RAM1, CPU_CLK_RAM1);
    });
    addOp("LSR", OP_CODE_INCR5, [&] {
        is.clock(1, Transition::SET_FLAG_RAM, CPU_CLK_RAM1, CPU_CLK_RAM1);
        is.clock(15, Transition::COPY_R1_R2, CPU_CLK_RAM1, CPU_CLK_RAMS);
        is.clock(1, Transition::DATA_0, 0, CPU_CLK_RAM2);
    });
    addOp("ROL", OP_CODE_INCR5, [&] {
        is.clock(1, Transition::COPY_FLAG, 0, CPU_CLK_RAM2);
        is.clock(15, Transition::COPY_R1_R2, CPU_CLK_RAM1, CPU_CLK_RAMS);
        is.clock(1, Transition::SET_FLAG_RAM, CPU_CLK_RAM1, CPU_CLK_RAM1);
    });
    addOp("ROR", 0, [&] {
        // Bit0 -> R2[VAL]
        copy_R1_to(CPU_CLK_RAM2, 1);
        is.clock(7, Transition::DATA_0, 0, CPU_CLK_RAM2);
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Stop
        setReadWrite(CPU_CLK_RAM2 | CPU_CLK_FLASH, 0); // R2/Flash read
        is.clock(7, 0, 0, CPU_CLK_RAM2 | CPU_CLK_FLASH); // Flash identity
        copy_R1_to(CPU_CLK_FLASH, 15); // R1[addr] -> Flash
        is.clock(1, Transition::COPY_FLAG, 0, CPU_CLK_FLASH); // Copy FLAG
        is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash identity enable
        is.clock(1, 0, CPU_CS_RAM2_FLASH, CPU_CLK_RAM2); // R1 stop
        is.clock(4, 0, 0, CPU_CLK_RAMS);
        is.clock(4, SPI_WRITE, 0, CPU_CLK_RAM1); // R1 write
        is.clock(4, RAM_ADDR_PTR, 0, CPU_CLK_RAM2); // PTR -> R2[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R2_to(CPU_CLK_RAM1, 16); // R2[PTR] -> R1[]
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[addr]
        is.addCmd(0, Transition::SET_FLAG_RAM); // Set flag according to R2
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_INCR5, CPU_CS_NONE);
    });
    addOp("ASR", OP_CODE_INCR5, [&] {
        is.clock(1, Transition::SET_FLAG_RAM, CPU_CLK_RAM1, CPU_CLK_RAM1);
        is.clock(14, Transition::COPY_R1_R2, CPU_CLK_RAM1, CPU_CLK_RAMS);
        is.clock(2, Transition::COPY_R1_R2, 0, CPU_CLK_RAM2);
        is.clock(1, Transition::DATA_0, CPU_CLK_RAM1, CPU_CLK_RAM1);
    });
    addOp("REV", 0, [&] {
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Stop
        setReadWrite(CPU_CLK_RAM2 | CPU_CLK_FLASH, 0); // R2/Flash read
        goFlashReverse(CPU_CLK_RAM2); // Flash reverse
        copy_R1_to(CPU_CLK_FLASH, 16); // R1[addr] -> Flash
        is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash reverse enable
        is.clock(1, 0, CPU_CS_RAM2_FLASH, CPU_CLK_RAM2); // R1 stop
        is.clock(5, 0, 0, CPU_CLK_RAMS);
        is.clock(3, SPI_WRITE, 0, CPU_CLK_RAM1); // R1 write
        is.clock(4, RAM_ADDR_PTR, 0, CPU_CLK_RAM2); // PTR -> R2[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R2_to(CPU_CLK_RAM1, 16); // R2[PTR] -> R1[]
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[addr]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_INCR5, CPU_CS_NONE);
    });
    
    // ADD|ADC|SUB|SBC|AND|OR|XOR (value):
    addOp("ADI", OP_CODE_INCR7, [&] {
        computeFlash(1, Transition::ADD_FLASH_first);
        computeFlash_(15, Transition::ADC_FLASH_tmp, Transition::ADC_FLASH);
    });
    addOp("ADCI", OP_CODE_INCR7, [&] {
        computeFlash_(16, Transition::ADC_FLASH_tmp, Transition::ADC_FLASH);
    });
    addOp("SUBI", OP_CODE_INCR7, [&] {
        computeFlash(1, Transition::SUB_FLASH_first);
        computeFlash_(15, Transition::SBC_FLASH_tmp, Transition::SBC_FLASH);
    });
    addOp("SBCI", OP_CODE_INCR7, [&] {
        computeFlash_(16, Transition::SBC_FLASH_tmp, Transition::SBC_FLASH);
    });
    addOp("ANDI", OP_CODE_INCR7, [&] {
        computeFlash(1, Transition::AND_FLASH_first);
        computeFlash(15, Transition::AND_FLASH);
    });
    addOp("ORI", OP_CODE_INCR7, [&] {
        computeFlash(1, Transition::OR_FLASH_first);
        computeFlash(15, Transition::OR_FLASH);
    });
    addOp("XORI", OP_CODE_INCR7, [&] {
        computeFlash(1, Transition::XOR_FLASH_first);
        computeFlash(15, Transition::XOR_FLASH);
    });
    
/******************************************************************************
 *
 *
 *
 ******************************************************************************/

    // Helper function:
    auto addOpExt = [&](const std::string& name,
                        std::function<void()> f) mutable {
        is.opcodeNext(name);
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Stop
        setReadWrite(CPU_CLK_RAM2 | CPU_CLK_FLASH, 0); // Flash read, R2 read
        is.clock(7, 0, 0, CPU_CLK_RAM2 | CPU_CLK_FLASH); // Flash identity
        copy_R1_to(CPU_CLK_FLASH, 16); // R1[addr] -> Flash
        is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash identity enable
        is.clock(1, 0, CPU_CS_RAM2_FLASH, CPU_CLK_RAM2); // R1 stop
        is.clock(4, 0, 0, CPU_CLK_RAMS);
        is.clock(4, SPI_WRITE, 0, CPU_CLK_RAM1); // R1 write
        is.clock(4, RAM_ADDR_PTR, 0, CPU_CLK_RAM2); // PTR -> R2[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R2_to(CPU_CLK_RAM1, 16); // R2[PTR] -> R1[]
        f(); // R2[VAL] & Flash -> R1[addr]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_INCR8, CPU_CS_NONE);
        is.fill();
    };
    
    // ADD|ADC|SUB|SBC|AND|OR|XOR (addr/addr):
    addOpExt("ADD", [&] {
        computeR2(1, Transition::ADD_FLASH_first);
        computeR2_(15, Transition::ADC_FLASH_tmp, Transition::ADC_FLASH);
    });
    addOpExt("ADC", [&] {
        computeR2_(16, Transition::ADC_FLASH_tmp, Transition::ADC_FLASH);
    });
    addOpExt("SUB", [&] {
        computeR2(1, Transition::SUB_FLASH_first);
        computeR2_(15, Transition::SBC_FLASH_tmp, Transition::SBC_FLASH);
    });
    addOpExt("SBC", [&] {
        computeR2_(16, Transition::SBC_FLASH_tmp, Transition::SBC_FLASH);
    });
    addOpExt("AND", [&] {
        computeR2(1, Transition::AND_FLASH_first);
        computeR2(15, Transition::AND_FLASH);
    });
    addOpExt("OR", [&] {
        computeR2(1, Transition::OR_FLASH_first);
        computeR2(15, Transition::OR_FLASH);
    });
    addOpExt("XOR", [&] {
        computeR2(1, Transition::XOR_FLASH_first);
        computeR2(15, Transition::XOR_FLASH);
    });

/******************************************************************************
 *
 *
 *
 ******************************************************************************/

    // JMP:
    is.opcodeNext("JMP");
    is.clock(1, SPI_WRITE, 0, CPU_CLK_RAM1);
    is.clock(16, RAM_ADDR_PC, 0, CPU_CLK_RAM1); // PC -> R1[]
    copy_Flash_to(CPU_CLK_RAM1, 24); // Flash -> R1[PC]
    is.addCmd(0, Transition::DATA_0); // Save write
    is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
    copy_pc();
    is.jump(OP_CODE_FETCH, CPU_CS_NONE);
    is.fill();
    
    // CALL:
    is.opcodeNext("CALL");
    is.clock(1, SPI_WRITE, 0, CPU_CLK_RAM1);
    is.clock(16, RAM_ADDR_PC, 0, CPU_CLK_RAM1); // PC -> R1[]
    copy_Flash_to(CPU_CLK_RAM1, 24); // Flash -> R1[PC]
    is.addCmd(0, Transition::DATA_0); // Save write
    is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
    setReadWrite(CPU_CLK_RAM2, CPU_CLK_RAM1); // R2 read, R1 write
    is.clock(12, 0, 0, CPU_CLK_RAMS);
    is.clock(4, RAM_ADDR_FLAG, 0, CPU_CLK_RAM2); // FLAG -> R2[]
    is.clock(4, RAM_ADDR_SPV, 0, CPU_CLK_RAM1); // SPV -> R1[]
    copy_R2_to(CPU_CLK_RAM1, 8); // R2[FLAG] -> R1[SPV]
    is.addValue(24, 4, CPU_CLK_RAM2, CPU_CLK_RAMS); // R2[PC] -> R1[SPV]
    is.addCmd(0, Transition::DATA_0); // Save write
    is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
    is.jump(OP_CODE_CPYPC, CPU_CS_NONE);
    is.fill();
    
    // RET:
    is.opcodeNext("RET");
    is.clock(1, SPI_READ, CPU_CS_RAM1, CPU_CLK_RAM1);
    is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Flash stop
    setReadWrite(0, CPU_CLK_RAM2); // R2 write
    is.clock(12, 0, 0, CPU_CLK_RAMS);
    is.clock(4, RAM_ADDR_FLAG, 0, CPU_CLK_RAM2); // FLAG -> R2[]
    is.clock(4, RAM_ADDR_SPV, 0, CPU_CLK_RAM1); // SPV -> R1[]
    is.addCmd(0, Transition::DATA_0); // Enable read
    copy_R1_to(CPU_CLK_RAM2, 32); // R1[SPV] -> R2[FLAG+PC]
    is.addCmd(0, Transition::DATA_0); // Save write
    is.jump(OP_CODE_FETCH, CPU_CS_NONE);
    is.fill();
    
    // BRFS|BRFC|BREQ|BRNE:
    auto brn = [&](const std::string& name, unsigned char mask, bool tstz)
        mutable {
        is.opcodeNext(name);
        if (tstz) {
            is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
            copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
            is.addCmd(0, Transition::DATA_0); // Enable read
            is.clock(1, Transition::TSTZ_RAM_first, CPU_CLK_RAM1, CPU_CLK_RAM1);
            is.clock(15, Transition::TSTZ_RAM, CPU_CLK_RAM1, CPU_CLK_RAM1);
            is.addCmd(0, Transition::DATA_0); // Buffer
        }
        is.setMask(mask);
        is.labelNext(name + " <- jump");
        if (tstz) {
            is.addCmd(CPU_CS_RAM2_FLASH, Transition::DATA_0); // Stop
            setReadWrite(0, CPU_CLK_RAM1); // R1 write
        } else
            is.clock(1, SPI_WRITE, 0, CPU_CLK_RAM1);
        is.clock(16, RAM_ADDR_PC, 0, CPU_CLK_RAM1); // PC -> R1[]
        copy_Flash_to(CPU_CLK_RAM1, 24); // Flash -> R1[PC]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_CPYPC, CPU_CS_NONE);
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Delimiter
        // Ends at 98 | 181
        is.setMask(~mask);
        is.labelNext(name + " <- skip1");
        if (!tstz) {
            is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        }
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // At 54 | 120
        is.repeat(tstz ? 61 : 44);
        incr_pure(tstz ? 6 : 4, false);
        is.jump(OP_CODE_FETCH, CPU_CS_NONE);
        is.fill();
    };
    brn("BRFS", Mask::MASK_FLAG1, false);
    brn("BRFC", Mask::MASK_FLAG0, false);
    brn("BREQ", Mask::MASK_FLAG1, true);
    brn("BRNE", Mask::MASK_FLAG0, true);

/******************************************************************************
 *
 *
 *
 ******************************************************************************/
 
    // LDI:
    auto ldi = [&](const std::string& name, size_t width,
                   unsigned char incrN) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_WRITE, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        copy_Flash_to(CPU_CLK_RAM1, width); // Flash -> R1[value]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(incrN, CPU_CS_NONE);
        is.fill();
    };
    ldi("LDI", 16, OP_CODE_INCR5);
    ldi("LDI32", 32, OP_CODE_INCR7);
    
    // LD:
    auto ld = [&](const std::string& name, size_t width) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM2, 16); // Flash -> R2[PTR]
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Enable read, Flash stop
        setReadWrite(CPU_CLK_FLASH, CPU_CLK_RAM2); // Flash read, R2 write
        goFlashReverse(CPU_CLK_RAM2); // Flash reverse
        copy_R1_to(CPU_CLK_FLASH, 16); // R1[addr] -> Flash
        is.clock(1, 0, 0, CPU_CLK_FLASH); // Flash reverse enable
        is.clock(1, 0, CPU_CS_RAM2_FLASH, CPU_CLK_RAM2); // R1 stop
        is.clock(5, 0, 0, CPU_CLK_RAMS);
        is.clock(3, SPI_READ, 0, CPU_CLK_RAM1); // R1 read
        is.clock(4, RAM_ADDR_VAL, 0, CPU_CLK_RAM2); // VAL -> R2[]
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R1_to(CPU_CLK_RAM2, width); // R1[addr] -> R2[VAL]
        if (width != 16)
            is.clock(16-width, 0, 0, CPU_CLK_RAM2); // 0 -> R2[VAL]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.jump(OP_CODE_CPY16, CPU_CS_NONE);
        is.fill();
    };
    ld("LD", 16);
    ld("LDB", 8);
    
    // ST:
    auto st = [&](const std::string& name, size_t width,
                  unsigned char jmp) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R1_to(CPU_CLK_RAM2, 16); // R1[addr] -> R2[PTR]
        is.addCmd(CPU_CS_RAM2_FLASH, Transition::DATA_0); // R1 stop
        setReadWrite(CPU_CLK_RAM1, 0); // R1 read
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(0, Transition::DATA_0); // Enable read
        copy_R1_to(CPU_CLK_RAM2, 16); // R1[addr] -> R2[VAL]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.jump(jmp, CPU_CS_NONE);
        is.fill();
    };
    st("ST", 16, OP_CODE_ST16);
    st("STB", 8, OP_CODE_ST8);
    
    // LD2:
    auto ld2 = [&](const std::string& name, size_t width) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Flash stop
        setReadWrite(CPU_CLK_FLASH | CPU_CLK_RAM2, 0); // Flash/R2 read
        goFlashReverse(0); // Flash reverse
        copy_R1_to(CPU_CLK_FLASH, 15); // R1[addr] -> Flash
        is.addCmd(CPU_CLK_RAM1, Transition::COPY_R1_R2);
        is.addCmd(CPU_CLK_FLASH | CPU_CS_RAM2_FLASH,
                  Transition::COPY_R1_R2); // Last clock R1 stop
        is.clock(1, 0, 0, CPU_CLK_FLASH | CPU_CLK_RAM1); // Flash enable
        is.clock(7, SPI_WRITE, 0, CPU_CLK_RAM1); // R1 write
        copy_Flash_to(CPU_CLK_RAM2, 16); // Flash -> R2[]
        is.clock(16, RAM_ADDR_SPV, 0, CPU_CLK_RAM1); // SPV -> R1[]
        copy_R2_to(CPU_CLK_RAM1, width); // R2[addr] -> R1[SPV]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.addCmd(CPU_CS_NONE, Transition::DATA_0); // Stop
        is.jump(OP_CODE_INCR3, CPU_CS_NONE);
        is.fill();
    };
    ld2("LD2W", 32);
    ld2("LD2", 16);
    
    // ST2:
    auto st2 = [&](const std::string& name, size_t width) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM1, 15); // Flash -> R1[]
        is.addCmd(CPU_CLK_FLASH, Transition::COPY_FLASH);
        is.addCmd(CPU_CLK_RAM1 | CPU_CS_RAM1,
                  Transition::COPY_FLASH); // Last clock Flash stop
        setReadWrite(CPU_CLK_FLASH, CPU_CLK_RAM2); // Flash read, R2 write
        goFlashReverse(0); // Flash reverse
        copy_R1_to(CPU_CLK_FLASH, 15); // R1[addr] -> Flash
        is.addCmd(CPU_CLK_RAM1, Transition::COPY_R1_R2);
        is.addCmd(CPU_CLK_FLASH | CPU_CS_RAM2_FLASH,
                  Transition::COPY_R1_R2); // Last clock R1 stop
        is.clock(1, 0, 0, CPU_CLK_FLASH | CPU_CLK_RAM1); // Flash enable
        is.clock(7, SPI_READ, 0, CPU_CLK_RAM1); // R1 read
        is.clock(16, RAM_ADDR_SPV, 0, CPU_CLK_RAM1); // SPV -> R1[]
        copy_Flash_to(CPU_CLK_RAM2, 16); // Flash -> R2[]
        copy_R1_to(CPU_CLK_RAM2, width); // R1[SPV] -> R2[]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.jump(OP_CODE_INCR3, CPU_CS_NONE);
        is.fill();
    };
    st2("ST2W", 32);
    st2("ST2", 16);
    
    // LPM:
    auto lpm = [&](const std::string& name, size_t w) mutable {
        is.opcodeNext(name);
        is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
        copy_Flash_to(CPU_CLK_RAM2, 16); // Flash -> R2[PTR]
        copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Stop
        setReadWrite(CPU_CLK_FLASH, CPU_CLK_RAM2); // Flash read, R2 write
        copy_R1_to(CPU_CLK_FLASH, 24); // R1[addr] -> Flash[VAL]
        is.clock(16, RAM_ADDR_VAL, 0, CPU_CLK_RAM2); // VAL -> R2[]
        copy_Flash_to(CPU_CLK_RAM2, w); // Flash[addr] -> R2[VAL]
        if (w != 16)
            is.clock(16-w, 0, 0, CPU_CLK_RAM2); // 0 -> R2[VAL]
        is.addCmd(0, Transition::DATA_0); // Save write
        is.jump(OP_CODE_CPY16, CPU_CS_NONE);
        is.fill();
    };
    lpm("LPM", 16);
    lpm("LPMB", 8);
    
    // OUT:
    is.opcodeNext("OUT");
    is.clock(1, SPI_READ, 0, CPU_CLK_RAM1);
    copy_Flash_to(CPU_CLK_RAM1, 16); // Flash -> R1[]
    is.addCmd(CPU_CS_RAM1, Transition::DATA_0); // Enable read
    {
        size_t cycles = 12;
        for (size_t i = 0; i < 16; ++i) {
            is.addCmd(CPU_CS_RAM1 | ((i&1) ? CPU_CLK_RAMS : 0),
                      Transition::COPY_R1_R2);
            is.repeat(cycles-1);
        }
        is.addCmd(CPU_CS_RAM1, Transition::DATA_0);
        is.repeat(cycles-1);
    }
    is.jump(OP_CODE_INCR3, CPU_CS_NONE);
    is.fill();
    
    // Check EPROM is filled
    check_eprom(eprom);
}

#endif
