
#include <iostream>
#include "instruction_set.h"
#include "instructions.h"
#include "eprom.h"
#include "hw.h"

FLASH flash("FLASH");
RAM ram1("RAM1");
RAM ram2("RAM2");
CPUeprom eprom;
HC595 hc595;
PCD8544 lcd;
OUT out;

int main()
{
    // Write instructions directly from code
    //flash.program(std::vector<unsigned char>({OP_CODE_HALT}));
    
    // Load instruction from file
    if (!flash.program("./_out.bin"))
        flash.program("../asm/_out.bin"); // Try different path
    
    // Construct an object with EPROM address lines
    CPUaddr addr(eprom);
    
    // Init EPROM with microinstructions
    std::cout << "\nEPROM INIT:\n";
    instruction_set_fill(eprom);
    if (eprom.error_occured)
        return 1; // Error
    
    // Generate binaries
    //eprom.write("_eprom.bin"); // EPROM microcode
    //flash.write("_flash.bin"); // FLASH lookups
    
    // Run simulator
    std::cout << "EPROM OK\n\nRUNNING:\n";
    lcd.init();
    while(1) {
    
        // Read EPROM data at current address
        const auto& data = addr.get();
        
        // Stop on HALT instruction
        if (!data.msg.empty()) {
            //if (data.msg[0] != '_')
            //    std::cout << data.msg << "\n";
            if (addr.addr.op == OP_CODE_HALT)
                break;
        }
        
        // Check if the EPROM has data configured on this address
        if (!data.configured) {
            std::cout << "\nUndefined data\n" << addr;
            break;
        }
        
        // Simulate 74HC595 with pin order: clk, din, rst, en
        auto op = hc595.step({data.flash, data.data, data.ram1, data.rst});
        addr.addr.op = op & ((1<<OP_BITS)-1);
        
        // Simulate memories with pin order: clk, din, cs
        auto f = flash.step({data.flash, data.data, data.cs_ram2_flash});
        auto r2 = ram2.step({data.ram2, data.data, data.cs_ram2_flash});
        auto r1 = ram1.step({data.ram1, data.data, data.cs_ram1});
        
        // Set input address lines (with virtual pull-downs)
        addr.addr.ram1_ram2 = (r1.out_1 || r2.out_1) ? 1 : 0;
        addr.addr.flash = (f.out_1) ? 1 : 0;
        addr.addr.flag = data.flag;
        
        // Check for data line collisions
        if (r1.out_0 + r1.out_1 + r2.out_0 + r2.out_1 > 1) {
            std::cout << "\nRam1/Ram2 collision\n" << addr;
            break;
        }
        if (f.out_0 + f.out_1 > 1) {
            std::cout << "\nFLASH collision\n" << addr;
            break;
        }
        
        // Print raw data output:
        //out.step({data.ram2, data.data, (unsigned char)((op>>7)&1)});
        
        // Simulate LCD with pin order: clk, din, cs, dc
        lcd.step({data.ram2, data.data, (unsigned char)((op>>7)&1),
                  (unsigned char)((op>>6)&1) });
        
        // Increment 74HC393 counter
        int cnt = (data.rst) ? 0 : ((int)addr.addr.cnt + 1);
        addr.addr.cnt = cnt;
        if (cnt >= 1<<CNT_BITS) {
            std::cout << "\nCNT overflow\n" << addr;
            break;
        }
    }
    
    // Mem dump (not implemented)
    std::cout << "\n\n";
}

